"""Massive :math:`F_2^{NC}` components."""

import LeProHQ
import numpy as np

from ..partonic_channel import RSL
from . import partonic_channel as pc
from .n3lo import interpolator


class GluonVV(pc.NeutralCurrentBase):
    """Vector-vector gluon component."""

    def NLO(self):
        """|ref| implements :eqref:`D.1`, :cite:`felix-thesis`."""

        def cg(z, _args):
            if self.is_below_pair_threshold(z):
                return 0.0
            return (
                self._FHprefactor / z * LeProHQ.cg0("F2", "VV", self._xi, self._eta(z))
            )

        return RSL(cg)

    def NNLO(self):
        """|ref| implements NLO (heavy) gluon coefficient function, :cite:`felix-thesis`."""

        def cg(z, _args):
            if self.is_below_pair_threshold(z):
                return 0.0
            return (
                self._FHprefactor
                / z
                * (4.0 * np.pi) ** 2
                * (
                    LeProHQ.cg1("F2", "VV", self._xi, self._eta(z))
                    + LeProHQ.cgBar1("F2", "VV", self._xi, self._eta(z))
                    * np.log(self._xi)
                )
            )

        return RSL(cg)

    def N3LO(self):
        """|ref| implements NNLO (heavy) gluon coefficient function, from N.Laurenti thesis."""

        coeff_iterpol = interpolator(
            "C2g", nf=self.nf, variation=self.n3lo_cf_variation
        )

        def cg(z, _args):
            if self.is_below_pair_threshold(z):
                return 0.0
            return coeff_iterpol(self._xi, self._eta(z))

        return RSL(cg)


class GluonAA(GluonVV):
    """Axial-vector-axial-vector gluon component."""

    def NLO(self):
        """|ref| implements :eqref:`D.4`, :cite:`felix-thesis`."""

        def cg(z, _args):
            if self.is_below_pair_threshold(z):
                return 0.0
            return (
                self._FHprefactor / z * LeProHQ.cg0("F2", "AA", self._xi, self._eta(z))
            )

        return RSL(cg)

    def NNLO(self):
        """|ref| implements NLO (heavy) gluon coefficient function, :cite:`felix-thesis`."""

        def cg(z, _args):
            if self.is_below_pair_threshold(z):
                return 0.0
            return (
                self._FHprefactor
                / z
                * (4.0 * np.pi) ** 2
                * (
                    LeProHQ.cg1("F2", "AA", self._xi, self._eta(z))
                    + LeProHQ.cgBar1("F2", "AA", self._xi, self._eta(z))
                    * np.log(self._xi)
                )
            )

        return RSL(cg)


class SingletVV(pc.NeutralCurrentBase):
    """Vector-vector singlet component."""

    def NNLO(self):
        """|ref| implements NLO (heavy) singlet coefficient function, :cite:`felix-thesis`."""

        def cq(z, _args):
            if self.is_below_pair_threshold(z):
                return 0.0
            return (
                self._FHprefactor
                / z
                * (4.0 * np.pi) ** 2
                * (
                    LeProHQ.cq1("F2", "VV", self._xi, self._eta(z))
                    + LeProHQ.cqBarF1("F2", "VV", self._xi, self._eta(z))
                    * np.log(self._xi)
                )
            )

        return RSL(cq)

    def N3LO(self):
        """|ref| implements NNLO (heavy) singlet coefficient function, from N.Laurenti thesis."""

        coeff_iterpol = interpolator(
            "C2q", nf=self.nf, variation=self.n3lo_cf_variation
        )

        def cq(z, _args):
            if self.is_below_pair_threshold(z):
                return 0.0
            return coeff_iterpol(self._xi, self._eta(z))

        return RSL(cq)


class SingletAA(pc.NeutralCurrentBase):
    """Axial-vector-axial-vector singlet component."""

    def NNLO(self):
        """|ref| implements NLO (heavy) singlet coefficient function, :cite:`felix-thesis`."""

        def cq(z, _args):
            if self.is_below_pair_threshold(z):
                return 0.0
            return (
                self._FHprefactor
                / z
                * (4.0 * np.pi) ** 2
                * (
                    LeProHQ.cq1("F2", "AA", self._xi, self._eta(z))
                    + LeProHQ.cqBarF1("F2", "AA", self._xi, self._eta(z))
                    * np.log(self._xi)
                )
            )

        return RSL(cq)


class NonSinglet(pc.NeutralCurrentBase):
    """Non-singlet, aka missing component."""

    def NNLO(self):
        """|ref| implements NLO (heavy) non-singlet coefficient function, :cite:`felix-thesis`."""

        def dq(z, _args):
            if self.is_below_pair_threshold(z):
                return 0.0
            # TODO move this hack into LeProHQ
            eta = self._eta(z)
            eta = min(eta, 1e8)
            r = (
                self._FHprefactor
                / z
                * (4.0 * np.pi) ** 2
                * LeProHQ.dq1("F2", "VV", self._xi, eta)
            )
            return r

        def Adler(_x, _args):
            # add minus sign
            return -LeProHQ.Adler("F2", "VV", self._xi)

        return RSL(dq, loc=Adler)



import numpy as np

from ..observable_name import ObservableName
from .esf import ESFInfo
from .result import ESFResult, EXSResult

GEV_CM2_CONV = 3.893793e10
"Conversion factor from GeV^-2 to 10^-38 cm^2"


def xs_coeffs_polarized(kind):
    """Compute coefficients in the definition of a given polarized cross section.

    Parameters
    ----------
    kind : str
        the identifier of the cross section for which coefficients have to be
        computed

    Returns
    -------
    np.ndarray
        the coefficients of the cross-section on the basis `(g4, gL, 2xg1)`
    """
    # derived SF, 2xg5
    if kind == "g5":
        return np.array([1.0, -1.0, 0.0])
    raise ValueError(f"'{kind}' cross section is not available.")


def xs_coeffs_unpolarized(kind, y, x=None, Q2=None, params=None):
    """Compute coefficients in the definition of a given unpolarized cross section.

    Parameters
    ----------
    kind : str
        the identifier of the cross section for which coefficients have to be
        computed
    y : float
        Bjorken y value for the chosen kinematic point
    x: None or float
        Bjorken x value for the chosen kinematic point
    Q2: None or float
        photon virtuality for the chosen kinematic point
    params : None or dict
        theory parameters required to fully specify the coefficients (which ones
        depends on the chosen definition, i.e. on `kind`)

    Returns
    -------
    np.ndarray
        the coefficients of the cross-section on the basis `(F2, FL, xF3)`

    Raises
    ------
    ValueError or KeyError
        in case not enough parameters have been specified

    """
    yp = 1.0 + (1.0 - y) ** 2
    ym = 1.0 - (1.0 - y) ** 2
    yL = y**2

    # derived SF, 2xF1
    if kind == "F1":
        return np.array([1.0, -1.0, 0.0])

    # Neutral Currents
    # HERA average
    if kind == "XSHERANCAVG":
        return np.array([1.0, -yL / yp, 0.0])

    if params is None:
        raise ValueError(f"Parameters required to compute '{kind}' cross section.")

    f3sign = -1 if params["projectilePID"] < 0 else 1
    # HERA fixed lepton
    if kind == "XSHERANC":
        return np.array([1.0, -yL / yp, f3sign * ym / yp])

    # Charged Currents
    norm = 0.0
    # HERA
    if kind == "XSHERACC":
        norm = 1.0 / 4.0
    # FW CDHSW see http://cds.cern.ch/record/200223/files/198909396.pdf
    elif kind == "FW":
        mn = np.sqrt(params["M2target"])
        norm = 1.0
        ym = 0
        yp = 1.0
        yL = y**2 / (2 * (y**2 / 2 + (1 - y) - (mn * x * y) ** 2 / Q2))
    elif kind == "XSFPFCC":
        INV_GEV_TO_PB = GEV_CM2_CONV / 100.0  # pb
        norm = (INV_GEV_TO_PB * params["GF"] ** 2) / (2.0 * np.pi)
        norm *= 1.0 / (2.0 * x * (1.0 + Q2 / params["M2W"]) ** 2)
    else:
        mn = np.sqrt(params["M2target"])
        m2w = params["M2W"]
        yp -= 2.0 * (mn * x * y) ** 2 / Q2  # = ypc
        # CHORUS
        if kind == "XSCHORUSCC":
            norm = (
                GEV_CM2_CONV
                * params["GF"] ** 2
                * mn
                / (2.0 * np.pi * (1.0 + Q2 / m2w) ** 2)
            )
        # NUTEV
        if kind == "XSNUTEVCC":
            norm = 100.0 / 2.0 / (1.0 + Q2 / m2w) ** 2
        # NUTEV neutrino dis
        if kind == "XSNUTEVNU":
            norm = GEV_CM2_CONV * params["GF"] ** 2 * mn / (2.0 * np.pi)
    return np.array([yp, -yL, f3sign * ym]) * norm


class EvaluatedCrossSection:
    def __init__(self, kin, obs_name, configs, get_esf):
        self.kin = kin
        self.Q2 = kin["Q2"]
        self.x = kin["x"]
        self.y = kin["y"]
        self.info = ESFInfo(obs_name, configs)
        self.get_esf = get_esf

    def alpha_qed_power(self):
        return 0

    def get_result(self):
        # Collect esfs
        flavor = self.info.obs_name.flavor

        # polarized ?
        if not self.info.obs_name.kind.startswith("g"):
            linear_coeffs = xs_coeffs_unpolarized(
                self.info.obs_name.kind,
                self.y,
                x=self.x,
                Q2=self.Q2,
                params=dict(
                    projectilePID=self.info.configs.coupling_constants.obs_config[
                        "projectilePID"
                    ],
                    M2target=self.info.configs.M2target,
                    M2W=self.info.configs.M2W,
                    GF=self.info.configs.GF,
                ),
            )
            sf1, sf2, sf3 = "F2", "FL", "F3"
        else:
            linear_coeffs = xs_coeffs_polarized(
                self.info.obs_name.kind,
            )
            sf1, sf2, sf3 = "g4", "gL", "g1"

        basic_sf1 = self.get_esf(
            ObservableName(f"{sf1}_{flavor}"), self.kin
        ).get_result()
        basic_sf2 = self.get_esf(
            ObservableName(f"{sf2}_{flavor}"), self.kin
        ).get_result()
        # skip F3, g1 if it is not required
        if linear_coeffs[2] != 0.0:
            basic_sf3 = self.get_esf(
                ObservableName(f"{sf3}_{flavor}"), self.kin
            ).get_result()
        else:
            basic_sf3 = ESFResult(self.x, self.Q2, basic_sf1.nf)

        # add normalizations
        esf = linear_coeffs @ np.array([basic_sf1, basic_sf2, basic_sf3])
        # remap to EXS
        sigma = EXSResult(self.x, self.Q2, self.y, basic_sf1.nf)
        for o, v in esf.orders.items():
            # now shift orders: push alpha_qed two powers up
            sigma.orders[(o[0], o[1] + self.alpha_qed_power(), o[2], o[3])] = v
        return sigma



r"""
The coefficient functions definition is given in :eqref:`4.2`,
:cite:`vogt-f2nc` (that is the main reference for their expression, i.e. all
the formulas in this module) (the same of :eqref:`1` in :cite:`vogt-flnc`).
The main reference for their expression is :cite:`vogt-flnc`.

Scale varitions main reference is :cite:`nnlo-sv-singlet` and
:cite:`nnlo-sv-nonsinglet`.

Scale variations main reference is :cite:`moch-f3nc`.

Note
----
    *   Check the theory reference for details on
        :doc:`../theory/scale-variations`

    *   The source files (available in fortran) follow the notations in :cite:`vogt-f3cc`
        where the odd-N moments are called minus even if they correspond to :math:`\nu + \bar{\nu}`.
        This convention is changed in :cite:`Davies:2016ruz` where the complete |N3LO| CC results are presented
        for the first time. Referred equations are not always in agreement with the code conventions.

    The code follows the notation:

    F3:
        * odd N: :math:`\nu + \bar{\nu}`, :math:`c_{ns,-}` common for CC and NC (up to N3LO).
        * even N: :math:`\nu - \bar{\nu}`, :math:`c_{ns,+} = \delta + c_{ns,-}` only for CC.
        * In :math:`c_{ns,+}` the term fl02 has to be turned off for CC.

    F2, FL:
        * odd N: :math:`\nu - \bar{\nu}`, :math:`c_{ns,-} = - \delta + c_{ns,+}` only for CC.
        * even N: :math:`\nu + \bar{\nu}`, :math:`c_{ns,+}` common for CC and NC (up to N3LO).
        * The term fl11 has to be turned off for CC.

"""

from . import kernels



from .. import partonic_channel as epc
from ..partonic_channel import RSL
from . import nlo, nnlo
from . import partonic_channel as pc


class NonSinglet(pc.LightBase):
    @staticmethod
    def LO():
        """|ref| Eq. (19) of :cite:`Borsa-light-nnlo-pol`."""
        return RSL.from_delta(1.0)

    @staticmethod
    def NLO():
        """|ref| Eq. (19) of :cite:`Borsa-light-nnlo-pol`."""
        return RSL.from_distr_coeffs(
            nlo.f2.ns_reg, (nlo.f2.ns_delta, nlo.f2.ns_omx, nlo.f2.ns_logomx)
        )

    def NNLO(self):
        """|ref| Eq. (19) of :cite:`Borsa-light-nnlo-pol`."""
        return RSL(
            nnlo.xc2ns2p.c2nn2a, nnlo.xc2ns2p.c2ns2b, nnlo.xc2ns2p.c2nn2c, [self.nf]
        )


class Gluon(epc.EmptyPartonicChannel):
    pass


class Singlet(epc.EmptyPartonicChannel):
    pass


class Valence(epc.EmptyPartonicChannel):
    pass



from .. import partonic_channel as epc
from ..partonic_channel import RSL
from . import nlo, nnlo
from . import partonic_channel as pc


# For gL, the LO contribution vanishes
class NonSinglet(pc.LightBase):
    @staticmethod
    def NLO():
        """
        |ref| implements NLO-part of :eqref:`A.1`, :cite:`Zijlstra-light-nnlo-pol`.
        """

        return RSL(nlo.fl.ns_reg)

    def NNLO(self):
        """
        |ref| implements :eqref:`19`, :cite:`Borsa-light-nnlo-pol`.
        """

        return RSL(
            nnlo.xclns2p.clnn2a, loc=nnlo.xclns2p.clnn2c, args=dict(reg=[self.nf])
        )


class Gluon(epc.EmptyPartonicChannel):
    pass


class Singlet(epc.EmptyPartonicChannel):
    pass


class Valence(epc.EmptyPartonicChannel):
    pass



import numba as nb
import numpy as np
from eko.constants import TR

from . import f2, f3


@nb.njit("f8(f8,f8[:])", cache=True)
def ns_reg(z, args):
    return f3.ns_reg(z, args)


# The coefficients are the same as
# F2LIGHT_NS
ns_delta = f2.ns_delta
ns_omx = f2.ns_omx
ns_logomx = f2.ns_logomx


@nb.njit("f8(f8,f8[:])", cache=True)
def gluon_reg(z, args):
    nf = args[0]
    return 4.0 * nf * TR * ((2.0 * z - 1.0) * np.log((1.0 - z) / z) - 4.0 * z + 3.0)



import numba as nb
import numpy as np
from eko import constants

from ... import special


@nb.njit("f8(f8,f8[:])", cache=True)
def pnsm_reg(z, args):
    """
    Regular :math:`P_{ns,-}` non-singlet valence-like splitting function (common
    to minus and valence).

    Parameters
    ----------
        z : float
            partonic momentum fraction
        args : np.ndarray
            further arguments

    Returns
    -------
        float
            the non-singlet valence-like splitting function :math:`P_{ns,-}(z)`

    """
    x = z
    NF = args[0]
    zeta2 = special.zeta2
    S2 = special.s2
    CF = constants.CF
    CA = constants.CA

    lnx = np.log(x)
    ln1mx = np.log(1 - x)
    pqq = 2 / (1 - x) - 1 - x
    pqqmx = 2 / (1 + x) - 1 + x
    S2x = S2(x)
    DM = 1 / (1 - x)

    # fmt: off
    GQQ1 = (
        2 * CF * NF * ( ( - 1.1111111111111112 - ( 2 * lnx )
        / 3 ) * pqq - ( 4 * ( 1 - x ) ) / 3 )
        + 4 * CA * CF * ( ( 3.7222222222222223 + ( 11 * lnx )
        / 6 + lnx**2 / 2 - np.pi**2 / 6 ) * pqq
        + ( 20 * ( 1 - x ) ) / 3 + lnx * ( 1 + x) )
        + 4 * CF**2 * ( ( ( - 3 * lnx ) / 2 - 2 * ln1mx
        * lnx ) * pqq - 5 * ( 1 - x ) - ( lnx**2 * ( 1
        + x ) ) / 2 - lnx * ( 1.5 + ( 7 * x ) / 2 ) )
        - 4 * CF * ( CF - CA / 2 ) * ( 2 * pqqmx * S2x
        + 4 * ( 1 - x ) + 2 * lnx * ( 1 + x ) )
    )
    # fmt: on

    #  The soft (`+'-distribution) part of the splitting function

    A2 = -40 / 9 * CF * NF + 268 / 9 * CA * CF - 8 * zeta2 * CA * CF

    GQQ1L = DM * A2

    #  The regular piece of the coefficient function

    X1NSMA = GQQ1 - GQQ1L

    return X1NSMA


@nb.njit("f8(f8,f8[:])", cache=True)
def pnsp_reg(z, args):
    """
    Regular :math:`P_{ns,+}` non-singlet singlet-like splitting function.

    Parameters
    ----------
        z : float
            partonic momentum fraction
        args : np.ndarray
            further arguments

    Returns
    -------
        float
            the non-singlet singlet-like splitting function :math:`P_{ns,+}(z)`

    """
    x = z
    NF = args[0]
    zeta2 = special.zeta2
    S2 = special.s2
    CF = constants.CF
    CA = constants.CA

    lnx = np.log(x)
    ln1mx = np.log(1 - x)
    pqq = 2 / (1 - x) - 1 - x
    pqqmx = 2 / (1 + x) - 1 + x
    S2x = S2(x)
    DM = 1 / (1 - x)

    # fmt: off
    GQQ1 = (
        2 * CF * NF * ( ( - 1.1111111111111112 - ( 2 * lnx )
        / 3 ) * pqq - ( 4 * ( 1 - x ) ) / 3 )
        + 4 * CA * CF * ( ( 3.7222222222222223 + ( 11 * lnx )
        / 6 + lnx**2 / 2 - np.pi**2 / 6 ) * pqq
        + ( 20 * ( 1 - x ) ) / 3 + lnx * ( 1 + x) )
        + 4 * CF**2 * ( ( ( - 3 * lnx ) / 2 - 2 * ln1mx
        * lnx ) * pqq - 5 * ( 1 - x ) - ( lnx**2 * ( 1
        + x ) ) / 2 - lnx * ( 1.5 + ( 7 * x ) / 2 ) )
        + 4 * CF * ( CF - CA / 2 ) * ( 2 * pqqmx * S2x
        + 4 * ( 1 - x ) + 2 * lnx * ( 1 + x ) )
    )
    # fmt: on

    #  The soft (`+'-distribution) part of the splitting function

    A2 = -40 / 9 * CF * NF + 268 / 9 * CA * CF - 8 * zeta2 * CA * CF

    GQQ1L = DM * A2

    #  The regular piece of the coefficient function

    X1NSPA = GQQ1 - GQQ1L

    return X1NSPA


@nb.njit("f8(f8,f8[:])", cache=True)
def pns_sing(z, args):
    """
    Singular :math:`P_{ns}` non-singlet splitting function (common to plus,
    minus, and valence).

    Parameters
    ----------
        z : float
            partonic momentum fraction
        args : np.ndarray
            further arguments

    Returns
    -------
        float
            the non-singlet splitting function :math:`P_{ns}(z)`

    """
    Y = z
    NF = args[0]
    zeta2 = special.zeta2
    CF = constants.CF
    CA = constants.CA

    A2 = -40 / 9 * CF * NF + 268 / 9 * CA * CF - 8 * zeta2 * CA * CF
    X1NSB = A2 / (1 - Y)

    return X1NSB


@nb.njit("f8(f8,f8[:])", cache=True)
def pns_loc(x, args):
    """
    Local :math:`P_{ns}` non-singlet splitting function (common to plus,
    minus, and valence).

    Parameters
    ----------
        x : float
            Bjorken :math:`x`
        args : np.ndarray
            further arguments

    Returns
    -------
        float
            the non-singlet splitting function :math:`P_{ns}(z)`

    """
    Y = x
    NF = args[0]
    zeta2 = special.zeta2
    zeta3 = special.zeta3
    CF = constants.CF
    CA = constants.CA

    # fmt: off
    P1DELT =(
        - 1/3*CF*NF
        + 3/2*CF**2
        + 17/6*CA*CF
        + 24*zeta3*CF**2
        - 12*zeta3*CA*CF
        - 8/3*zeta2*CF*NF
        - 12*zeta2*CF**2
        + 44/3*zeta2*CA*CF
    )
    # fmt: on

    #  The soft (`+'-distribution) part of the splitting function

    A2 = -40 / 9 * CF * NF + 268 / 9 * CA * CF - 8 * zeta2 * CA * CF

    X1NSC = np.log(1 - Y) * A2 + P1DELT

    return X1NSC



"""Module controlling the log outputs."""

import logging
import os

from rich.logging import RichHandler

# read environment
log_level = int(os.environ.get("YADISM_LOG_LEVEL", logging.INFO))
log_to_stdout = bool(os.environ.get("YADISM_LOG_STDOUT", True))
log_file = os.environ.get("YADISM_LOG_FILE")
silent_mode = bool(os.environ.get("YADISM_SILENT_MODE", False))

debug = bool(os.environ.get("DEBUG", False))

module_name = __name__.split(".")[0]
logger = logging.getLogger(module_name)
ekologger = logging.getLogger("eko")


def setup(
    console=None,
    log_level=log_level,
    log_to_stdout=log_to_stdout,
    log_file=log_file,
):
    """Initialize log outputs.

    Parameters
    ----------
    console: rich.console.Console
        rich high level console interface
    log_level: str
        logging level
    log_to_stdout: boolean-like
        switch on/off printing on STD output
    log_file: str-like
        redirect output logs to file

    """
    logger.setLevel(log_level)
    ekologger.setLevel(log_level)
    logger.handlers = []
    ekologger.handlers = []

    # add rich logger
    if log_to_stdout:
        rh = RichHandler(log_level, console=console)
        rh.setFormatter(logging.Formatter("%(message)s", datefmt="[%X]"))
        logger.addHandler(rh)
        ekologger.addHandler(rh)
    # add file logger
    if log_file is not None:
        fh = logging.FileHandler(log_file)
        logger.addHandler(fh)
        ekologger.addHandler(fh)



"""Collect all kernels for given |FNS|."""

import numpy as np
from eko.matchings import nf_default

from . import asy, heavy, intrinsic, kernels, light
from .partonic_channel import EmptyPartonicChannel


class Component(list):
    """Used for organize elements and debugging purpose."""

    heavyness = {0: "light", 4: "charm", 5: "bottom", 6: "top"}

    def __init__(self, heavy, kernels=None):
        self.heavy = self.heavyness[heavy]

        if kernels is None:
            kernels = []
        super().__init__(kernels)

    def __repr__(self):
        return self.heavy + f"({len(self)} kernels)"


# TODO add more doc strings


class Combiner:
    """Does the matching between coefficient functions and partons with their appropriate coupling strength.

    Parameters
    ----------
    esf : EvaluatedStructureFunction
        current ESF

    """

    def __init__(self, esf):
        self.esf = esf
        self.masses = {4 + i: not mass for i, mass in enumerate(esf.info.ZMq)}
        self.obs_name = esf.info.obs_name
        self.nf = nf_default(esf.Q2, esf.info.threshold)
        self.target = esf.info.target
        self.scheme = esf.info.scheme
        self.fonllparts = esf.info.fonllparts

    def collect(self):
        """Collect all kernels."""
        comps = []
        family = self.obs_name.flavor_family
        # Adding light component
        if family in ["light", "total"] and self.fonllparts in ["massless", "full"]:
            comps.append(self.light_component())
        if family == "heavy" and self.fonllparts in ["massless", "full"]:
            # the only case in which an heavy contribution is not present in those
            # accounted for in total, it's when heavy already became heavylight
            comps.extend(self.heavylight_components())
        if family in ["heavy", "total"] and self.fonllparts in ["massive", "full"]:
            comps.extend(self.heavy_components())
        return comps

    def light_component(self):
        """Collect massless kernels."""
        nf = self.nf
        masses = self.masses

        comp = Component(0)
        # light does not contain any mass effects and so is just always there
        comp.extend(light.kernels.generate(self.esf, nf))
        # instead missing encodes mass effects and so we need to fork:
        for ihq in range(nf + 1, 7):
            if masses[ihq]:
                if "FFN0" in self.scheme:
                    comp.extend(
                        asy.kernels.generate_missing_asy(
                            self.esf,
                            nf,
                            ihq,
                            self.esf.info.theory["pto_evol"],
                        )
                    )
                else:
                    comp.extend(heavy.kernels.generate_missing(self.esf, nf, ihq))

        return comp

    def heavylight_components(self):
        """Collect single-flavor massless kernels."""
        nf = self.nf
        hq = self.obs_name.hqnumber
        masses = self.masses

        comps = []

        if hq < nf or (hq == nf and not masses[hq]):
            heavylight = Component(hq)
            heavylight.extend(kernels.generate_single_flavor_light(self.esf, nf, hq))
            comps.append(heavylight)

        return comps

    def heavy_components(self):
        """Collect massive kernels."""
        nf = self.nf
        hq = self.obs_name.hqnumber
        masses = self.masses

        comps = []

        heavy_comps = {}
        # The loop starts at nf because nf counts the number of quarks of which
        # are above the mass threshold. For these quarks the masses are not
        # considered.
        for sfh in range(nf, 7):
            # exclude sfh=3, since heavy contributions are there for [4,5,6]
            # if it's ZM you don't even have the component
            if sfh not in masses:
                continue

            # There is no massive heavy contribution for ZM
            if not masses[sfh]:
                continue

            heavy_comps[sfh] = Component(sfh)

            # calculate only the contribution corresponding to the observable
            # i.e. charm for F_charm, bottom for F_bottom. In the case of
            # F_total (if hq=0), sum over all massive contributions.
            if hq not in (0, sfh):
                continue

            # heavy quark is intrinsic
            if "FFN0" in self.scheme:
                heavy_comps[sfh].extend(
                    asy.kernels.generate_intrinsic_asy(
                        self.esf, nf, self.esf.info.theory["pto_evol"], ihq=sfh
                    ),
                )
            else:
                heavy_comps[sfh].extend(intrinsic.kernels.generate(self.esf, ihq=sfh))

            if "FFN0" in self.scheme:
                heavy_comps[sfh].extend(
                    asy.kernels.generate_heavy_asy(
                        self.esf, nf, self.esf.info.theory["pto_evol"], ihq=sfh
                    )
                )
            else:
                heavy_comps[sfh].extend(heavy.kernels.generate(self.esf, nf, ihq=sfh))

            comps.append(heavy_comps[sfh])
        return comps

    @staticmethod
    def apply_isospin(full, z, a):
        """Apply isospin symmetry to u and d distributions.

        Parameters
        ----------
        full : list(yadism.kernels.Kernel)
            all participants
        z : float
            number of protons
        a : float
            atomic mass number

        """
        nucl_factors = np.array([[z, a - z], [a - z, z]]) / a
        for ker in full:
            for sign in [-1, 1]:
                ps = np.array(
                    [ker.partons.get(sign * 1, 0), ker.partons.get(sign * 2, 0)]
                )
                ker.partons[sign * 1], ker.partons[sign * 2] = nucl_factors @ ps

    @staticmethod
    def drop_empty(full):
        """Drop kernels with :class:`EmptyPartonicChannel` or its partons with empty weight.

        Parameters
        ----------
        elems : list(yadism.kernels.Kernel)
            all participants

        Returns
        -------
        filtered_kernels : list(yadism.kernels.Kernel)
            improved participants

        """
        filtered_kernels = []
        for ker in full:
            ker.partons = {p: w for p, w in ker.partons.items() if w != 0}
            if len(ker.partons) > 0 and not isinstance(ker.coeff, EmptyPartonicChannel):
                filtered_kernels.append(ker)
        return filtered_kernels

    def collect_elems(self):
        """Collect all kernels according to the |FNS|.

        Returns
        -------
        elems : list(yadism.kernels.Kernel)
            all participants

        """
        components = self.collect()

        full = []
        for comp in components:
            full.extend(comp)

        # add level-0 nuclear correction: apply isospin symmetry
        self.apply_isospin(full, self.target["Z"], self.target["A"])

        # drop all kernels with 0 weight, or empty coeffs
        return self.drop_empty(full)



from ..partonic_channel import RSL
from . import nlo, nnlo
from . import partonic_channel as pc


class NonSinglet(pc.LightBase):
    @staticmethod
    def LO():
        """
        |ref| implements LO-part of :eqref:`A.1`, :cite:`Zijlstra-light-nnlo-pol`.
        """

        # leading order is just a delta function
        return RSL.from_delta(1.0)

    @staticmethod
    def NLO():
        """
        |ref| implements the equivalence relation between :math:`g_1` and :math:`F_3`
        as explained in :eqref:`A.19` (and paragraph below) of :cite:`Borsa-light-nnlo-pol`.
        For the explicit expressions, refer to :eqref:`A.1` of :cite:`Zijlstra-light-nnlo-pol`
        and :eqref:`15` of :cite:`deFlorian-light-nlo-pol`.
        """

        return RSL.from_distr_coeffs(
            nlo.g1.ns_reg, (nlo.g1.ns_delta, nlo.g1.ns_omx, nlo.g1.ns_logomx)
        )

    def NNLO(self):
        """
        |ref| implements the equivalence relation between :math:`g_1` and :math:`F_3`
        as explained in :eqref:`A.19` (and paragraph below) of :cite:`Borsa-light-nnlo-pol`.
        For the explicit expressions, refer to :eqref:`A.1` of :cite:`Zijlstra-light-nnlo-pol`
        and :eqref:`15` of :cite:`deFlorian-light-nlo-pol`.
        """

        return RSL(
            nnlo.xc3ns2p.c3nm2a, nnlo.xc3ns2p.c3ns2b, nnlo.xc3ns2p.c3nm2c, [self.nf]
        )


class Gluon(pc.LightBase):
    def NLO(self):
        r"""
        |ref| implements NLO-part of :eqref:`A.5`, :cite:`Zijlstra-light-nnlo-pol`.
        For a simpler expression, refer also to :eqref:`9` of :cite:`deFlorian-light-nlo-pol`.

        Note
        ----
        2 * n_f is coming from momentum sum rule q_i -> {q_i, g} but
        g -> {g, q_i, \bar{q_i} forall i}, so the 2 * n_f is needed
        to compensate for all the number of flavours plus antiflavours
        in which the gluon can go.
        """

        return RSL(nlo.g1.gluon_reg, args=[self.nf])

    def NNLO(self):
        """
        |ref| implements |NNLO| massless contribution of :eqref:`A.5`, :cite:`Zijlstra-light-nnlo-pol`.

        Note
        ----
        The actual implementation is taken from Apfel++ :cite:`Bertone:2017gds`:, see also:

            https://github.com/vbertone/apfelxx/blob/master/src/structurefunctions/zeromasscoefficientfunctionspol_sl.cc

        """
        return RSL(nnlo.g1.gluon_reg, args=[self.nf])


class Singlet(pc.LightBase):
    def NNLO(self):
        """
        |ref| implements |NNLO| massless contribution of :eqref:`A.4`, :cite:`Zijlstra-light-nnlo-pol`.

        Note
        ----
        The actual implementation is taken from Apfel++ :cite:`Bertone:2017gds`:, see also:

            https://github.com/vbertone/apfelxx/blob/master/src/structurefunctions/zeromasscoefficientfunctionspol_sl.cc

        """
        return RSL(nnlo.g1.singlet_reg, args=[self.nf])



"""Module that computes a physical observable.

The observable could be structure functions or some linear
combination of them.

"""

import logging

from .esf import exs

logger = logging.getLogger(__name__)


class CrossSection:
    r"""Represents an abstract structure function.

    This class acts as an intermediate handler between the :py:class:`Runner`
    exposed to the outside and the :py:class:`EvaluatedStructureFunction`
    which compute the actual observable.

    Parameters
    ----------
    obs_name : ObservableName
        name
    runner : yadism.runner.Runner
        parent reference
    """

    def __init__(self, obs_name, runner):
        # internal managers
        self.obs_name = obs_name
        self.runner = runner
        self.exss = []
        logger.debug("Init %s", self)

    def __repr__(self):
        return self.obs_name.name

    def __len__(self):
        return len(self.exss)

    @property
    def elements(self):
        """Collect the evaluated observable."""
        return self.exss

    def load(self, kinematic_configs):
        r"""Load all kinematic configurations from the run card.

        Parameters
        ----------
        kinematic_configs : list(dict)
            run card input

        """
        self.exss = []
        for kinematics in kinematic_configs:
            self.exss.append(
                exs.EvaluatedCrossSection(
                    kinematics, self.obs_name, self.runner.configs, self.get_esf
                )
            )

    def get_esf(self, obs_name, kin):
        r"""Get the observable in question for given kinematics.

        Parameters
        ----------
        obs_name: str
            name of the observable
        kin: dict
            contains the kinematic specs

        Returns
        -------
        get_sf:
            observable evaluated on some kinematics

        """
        return self.runner.get_sf(obs_name).get_esf(obs_name, kin, use_raw=False)

    def get_result(self):
        r"""Collect the results from all childrens.

        Returns
        -------
        output : list(ESFResult)
            all children outputs

        """
        return list(elem.get_result() for elem in self.elements)



import LeProHQ
import numpy as np

from ..partonic_channel import RSL
from . import partonic_channel as pc


class NonSinglet(pc.NeutralCurrentBase):
    def NNLO(self):
        """
        |ref| implements |NNLO| (heavy) NS coefficient function,
        :eqref:`D.58` of :cite:`felix-thesis`.
        """

        def dq(z, _args):
            if self.is_below_pair_threshold(z):
                return 0.0
            eta = self._eta(z)
            eta = min(eta, 1e5)
            return (
                self._FHprefactor
                / z
                * (4.0 * np.pi) ** 2
                * LeProHQ.dq1("g4", "VA", self._xi, eta)
            )

        def Adler(_x, _args):
            # add minus sign
            return -LeProHQ.Adler("g4", "VA", self._xi)

        return RSL(dq, loc=Adler)



import numpy as np
from eko import basis_rotation as br

from .. import kernels


def import_pc_module(kind, process):
    return kernels.import_local(kind, process, __name__)


def generate(esf, nf):
    """
    Collect the light coefficient functions

    Parameters
    ----------
        esf : EvaluatedStructureFunction
            kinematic point
        nf : int
            number of light flavors

    Returns
    -------
        elems : list(yadism.kernels.Kernel)
            list of elements
    """
    kind = esf.info.obs_name.kind
    pcs = import_pc_module(kind, esf.process)
    coupling = esf.info.coupling_constants
    is_pv = esf.info.obs_name.is_parity_violating
    if esf.process == "CC":
        weights_even = kernels.cc_weights_even(
            coupling,
            esf.Q2,
            br.quark_names[:nf],
            nf,
            is_pv,
        )
        ns_even = kernels.Kernel(weights_even["ns"], pcs.NonSingletEven(esf, nf))
        weights_odd = kernels.cc_weights_odd(
            esf.info.coupling_constants,
            esf.Q2,
            br.quark_names[:nf],
            nf,
            is_pv,
        )
        ns_odd = kernels.Kernel(weights_odd["ns"], pcs.NonSingletOdd(esf, nf))

        if is_pv:
            v = kernels.Kernel(weights_odd["v"], pcs.Valence(esf, nf))
            return (ns_even, ns_odd, v)
        g = kernels.Kernel(weights_even["g"], pcs.Gluon(esf, nf))
        s = kernels.Kernel(weights_even["s"], pcs.Singlet(esf, nf))
        return (ns_even, g, s, ns_odd)

    # NC standard weights
    weights = nc_weights(
        esf.info.coupling_constants,
        esf.Q2,
        nf,
        is_pv,
    )

    # let's separate according to parity
    if is_pv:
        ns = kernels.Kernel(
            weights["ns"],
            pcs.NonSinglet(esf, nf),
        )
        v = kernels.Kernel(
            weights["v"],
            pcs.Valence(esf, nf),
        )
        return [ns, v]

    ns = kernels.Kernel(
        weights["ns"],
        pcs.NonSinglet(esf, nf),
    )
    g = kernels.Kernel(
        weights["g"],
        pcs.Gluon(esf, nf),
    )
    s = kernels.Kernel(
        weights["s"],
        pcs.Singlet(esf, nf),
    )
    kernels_list = [ns, g, s]

    # at N3LO we need to add also the fl11 diagrams
    if esf.info.theory["pto"] == 3:

        gluon_fl11 = pcs.GluonFL11(esf, nf)
        quark_fl11 = pcs.QuarkFL11(esf, nf)
        weights_fl11 = nc_fl11_weights(esf.info.coupling_constants, esf.Q2, nf)
        ns_fl11 = kernels.Kernel(
            weights_fl11["q"],
            quark_fl11,
        )
        g_fl11 = kernels.Kernel(weights_fl11["g"], gluon_fl11)
        kernels_list.extend([ns_fl11, g_fl11])
    return kernels_list


def nc_weights(coupling_constants, Q2, nf, is_pv, skip_heavylight=False):
    """
    Compute light NC weights.

    Parameters
    ----------
    coupling_constants : CouplingConstants
        manager for coupling constants
    Q2 : float
        W virtuality
    nf : int
        number of light flavors
    is_pv: bool
        True if observable violates parity conservation
    skip_heavylight : bool
        prevent the last quark to couple to the boson


    Returns
    -------
    weights : dict
        mapping pid -> weight for ns, g and s channel
    """
    # quark couplings
    tot_ch_sq = 0
    ns_partons = {}
    pids = range(1, nf + 1)

    for q in pids:
        # we don't want this quark to couple to the photon (because it is treated separately),
        # but still let it take part in the average
        if skip_heavylight and q == nf:
            continue
        if is_pv:
            w = coupling_constants.get_weight(
                q, Q2, "VA"
            ) + coupling_constants.get_weight(q, Q2, "AV")
        else:
            w = coupling_constants.get_weight(
                q, Q2, "VV"
            ) + coupling_constants.get_weight(q, Q2, "AA")
        ns_partons[q] = w
        ns_partons[-q] = w if not is_pv else -w
        tot_ch_sq += w

    # compute gluon, singlet or valence
    ch_av = tot_ch_sq / len(pids)
    if is_pv:
        v_partons = {q: np.sign(q) * ch_av for q in [*pids, *(-q for q in pids)]}
        return {"ns": ns_partons, "v": v_partons}

    # gluon and singlet coupling = charge average (omitting the *2/2)
    s_partons = {q: ch_av for q in [*pids, *(-q for q in pids)]}
    return {"ns": ns_partons, "g": {21: ch_av}, "s": s_partons}


def nc_fl11_weights(coupling_constants, Q2, nf, skip_heavylight=False):
    """Compute the NC weights for the flavor class :math:`fl_{11}`.

    For the time being we don't have such diagrams for parity violating
    structure functions.

    Parameters
    ----------
    coupling_constants : CouplingConstants
        manager for coupling constants
    Q2 : float
        W virtuality
    nf : int
        number of light flavors
    skip_heavylight : bool
        prevent the last quark to couple to the boson

    Returns
    -------
    weights : dict
        mapping pid -> weight for ns, g and s channel
    """
    # quark couplings
    quark_partons = {}
    tot_ch_sq = 0
    pids = range(1, nf + 1)

    for q in pids:
        # we don't want this quark to couple to the photon (because it is treated separately),
        # but still let it take part in the average
        if skip_heavylight and q == nf:
            continue
        w = coupling_constants.get_fl11_weight(
            q, Q2, nf, "VV"
        ) + coupling_constants.get_fl11_weight(q, Q2, nf, "AA")
        quark_partons[q] = w
        quark_partons[-q] = w
        tot_ch_sq += w

    # compute gluon
    ch_av = tot_ch_sq / len(pids)

    # NOTE: since here we are implementing the plain coupling
    # Q1 * Tr[Q2] / nf, we have not decomposed the quark sector
    # into singlet and non singlet for this class of diagrams
    return {"q": quark_partons, "g": {21: ch_av}}



from eko import basis_rotation as br

from .. import kernels


def import_pc_module(kind, process):
    return kernels.import_local(kind, process, __name__)


def generate(esf, ihq):
    """
    Collect the light coefficient functions

    Parameters
    ----------
        esf : EvaluatedStructureFunction
            kinematic point
        ihq : int
            intrinsic quark

    Returns
    -------
        elems : list(yadism.kernels.Kernel)
            list of elements
    """

    kind = esf.info.obs_name.kind
    is_pv = esf.info.obs_name.is_parity_violating
    cfs = import_pc_module(kind, esf.process)
    m2hq = esf.info.m2hq[ihq - 4]
    if esf.process == "CC":
        w = kernels.cc_weights(
            esf.info.coupling_constants,
            esf.Q2,
            br.quark_names[ihq - 1],
            ihq,
            is_pv,
        )
        wq = {k: v for k, v in w["ns"].items() if abs(k) == ihq}
        if is_pv:
            return (kernels.Kernel(wq, cfs.Rplus(esf, ihq - 1, m1sq=m2hq)),)
        return (kernels.Kernel(wq, cfs.Splus(esf, ihq - 1, m1sq=m2hq)),)

    if is_pv:
        wVA = esf.info.coupling_constants.get_weight(ihq, esf.Q2, "VA")
        wAV = esf.info.coupling_constants.get_weight(ihq, esf.Q2, "AV")
        wp = wVA + wAV
        wm = wVA - wAV
        return (
            kernels.Kernel(
                {ihq: wp, (-ihq): -wp},
                cfs.Rplus(esf, ihq - 1, m1sq=m2hq, m2sq=m2hq),
            ),
            kernels.Kernel(
                {ihq: wm, (-ihq): -wm},
                cfs.Rminus(esf, ihq - 1, m1sq=m2hq, m2sq=m2hq),
            ),
        )
    wVV = esf.info.coupling_constants.get_weight(ihq, esf.Q2, "VV")
    wAA = esf.info.coupling_constants.get_weight(ihq, esf.Q2, "AA")
    wp = wVV + wAA
    wm = wVV - wAA
    return (
        kernels.Kernel(
            {ihq: wp, (-ihq): wp}, cfs.Splus(esf, ihq - 1, m1sq=m2hq, m2sq=m2hq)
        ),
        kernels.Kernel(
            {ihq: wm, (-ihq): wm}, cfs.Sminus(esf, ihq - 1, m1sq=m2hq, m2sq=m2hq)
        ),
    )



# auto-generated module by ic package
# fmt: off
# pylint: skip-file

import numpy as np
from scipy.special import spence


def li2(x):
    return spence(1-x)

def f1_splus_raw(pc):
    return -(pc.L_xi*(-4*pc.m1sq*pc.m2sq*pc.s1hat**2 + pc.s1hat**3*(pc.s1hat + 4*pc.sigma_pm) + pc.s1hat*(2*pc.delta**2 + 7*pc.s1hat*pc.sigma_pm)*pc.sigma_pp + 2*(pc.delta**2 + 2*pc.s1hat*pc.sigma_pm)*pc.sigma_pp**2))/(2.*pc.deltap**3*pc.s1hat) + (-2*pc.delta**2*(9*pc.s1hat**3 + 4*pc.m2sq**2*pc.sigma_pp + 2*pc.s1hat*(4*pc.m2sq + pc.s1hat)*pc.sigma_pp) + pc.s1hat*(pc.s1hat*(28*pc.m1sq**2*pc.s1hat + pc.s1hat**3 + 2*pc.m1sq*pc.s1hat*(-8*pc.m2sq + 14*pc.Q2 + 3*pc.s1hat) - 3*pc.s1hat**2*pc.sigma_pm + 4*pc.m2sq*pc.sigma_mp*pc.sigma_pm) + (4*pc.m1sq*pc.m2sq*pc.s1hat + 7*pc.s1hat**2*pc.sigma_mp - 8*pc.m2sq*(2*pc.m2sq + 5*pc.s1hat)*pc.sigma_pm)*pc.sigma_pp - 3*pc.s1hat*pc.sigma_mp*pc.sigma_pp**2))/(4.*pc.deltap**2*pc.s1hat*(pc.m2sq + pc.s1hat)**2)

def f1_splus_soft(pc):
    return -((pc.sigma_pp*(2*pc.delta + pc.L_xisoft*pc.sigma_pp))/pc.delta**2)

def f1_splus_virt(pc):
    return (-2*pc.Cplus*np.sqrt(pc.m1sq * pc.m2sq) + pc.CRm*pc.sigma_pp)/(2.*pc.delta)

def f1_sminus_raw(pc):
    return (2*pc.L_xi*np.sqrt(pc.m1sq * pc.m2sq)*(pc.delta**2*pc.sigma_pp + pc.s1hat*(pc.deltap**2 + (pc.m2sq + pc.Q2)*pc.s1hat + 2*pc.sigma_pm*pc.sigma_pp)))/(pc.deltap**3*pc.s1hat) + (np.sqrt(pc.m1sq * pc.m2sq)*(2*pc.delta**2*(2*pc.m2sq + pc.s1hat) + pc.s1hat*(2*pc.deltap**2 + 8*pc.m2sq*pc.sigma_pm + pc.s1hat*(pc.s1hat - 2*pc.sigma_mp + 4*pc.sigma_pm + pc.sigma_pp))))/(pc.deltap**2*pc.s1hat*(pc.m2sq + pc.s1hat))

def f1_sminus_soft(pc):
    return (2*np.sqrt(pc.m1sq * pc.m2sq)*(2*pc.delta + pc.L_xisoft*pc.sigma_pp))/pc.delta**2

def f1_sminus_virt(pc):
    return (-2*pc.CRm*np.sqrt(pc.m1sq * pc.m2sq) + pc.Cplus*pc.sigma_pp)/(2.*pc.delta)

def f2_splus_raw(pc):
    return (pc.s1hat*(-2*(pc.m1sq + pc.m2sq)*pc.s1hat**2 - 2*(pc.delta**2 - 6*pc.m1sq*pc.Q2)*(pc.m2sq + pc.s1hat) - (2*pc.delta**4*(2*pc.m2sq + pc.s1hat))/pc.s1hat**2 - 9*pc.m2sq*pc.sigma_pm**2 + 2*pc.s1hat*(2*pc.delta**2 + (pc.m1sq - 5*pc.m2sq)*pc.sigma_pm) - (2*pc.delta**2*(pc.delta**2 + 2*(2*pc.m2sq + pc.s1hat)*pc.sigma_pm))/pc.s1hat - pc.delta**2*(pc.m2sq - 2*pc.sigma_pp) + ((pc.deltap**2 - 6*pc.Q2*(pc.m2sq + pc.s1hat))*pc.sigma_pp*(pc.s1hat + pc.sigma_pp))/(2.*(pc.m2sq + pc.s1hat)))*pc.x)/(pc.deltap**2*pc.Q2*(pc.m2sq + pc.s1hat)) - (pc.L_xi*(2*pc.delta**4*(pc.s1hat + pc.sigma_pp) + 2*pc.delta**2*pc.s1hat*(pc.s1hat + 2*pc.sigma_pm)*(pc.s1hat + pc.sigma_pp) + pc.s1hat**2*(pc.deltap**2*pc.sigma_pp - 6*pc.m1sq*pc.Q2*(2*pc.s1hat + 3*pc.sigma_pp)))*pc.x)/(pc.deltap**3*pc.Q2*pc.s1hat)

def f2_splus_soft(pc):
    return (-2*(2*pc.delta + pc.L_xisoft*pc.sigma_pp)*pc.x)/pc.Q2

def f2_splus_virt(pc):
    return (pc.delta*(2*pc.CRm + pc.C1p*pc.m1sq + pc.C1m*pc.m2sq)*pc.x)/(2.*pc.Q2)

def f2_sminus_raw(pc):
    return (2*pc.L_xi*np.sqrt(pc.m1sq * pc.m2sq)*(pc.deltap**2 - 6*pc.m1sq*pc.Q2)*pc.s1hat*pc.x)/(pc.deltap**3*pc.Q2) - (np.sqrt(pc.m1sq * pc.m2sq)*pc.s1hat*(pc.deltap**2*(-4*pc.m2sq - 3*pc.s1hat + pc.sigma_pp) + 6*pc.Q2*(pc.m2sq + pc.s1hat)*(pc.s1hat + pc.sigma_pp))*pc.x)/(pc.deltap**2*pc.Q2*(pc.m2sq + pc.s1hat)**2)

def f2_sminus_soft(pc):
    return 0

def f2_sminus_virt(pc):
    return (pc.delta*(2*pc.Cplus + (pc.C1m + pc.C1p)*np.sqrt(pc.m1sq * pc.m2sq))*pc.x)/(2.*pc.Q2)

def fl_splus_raw(pc):
    return -((8*pc.delta**4*(pc.m2sq + pc.s1hat)**2 + 2*pc.delta**2*(pc.s1hat*(pc.s1hat*(3*pc.m2sq**2 + pc.m2sq*pc.s1hat - pc.s1hat*(9*pc.Q2 + 2*pc.s1hat)) + 4*(pc.m2sq + pc.s1hat)*(2*pc.m2sq + pc.s1hat)*pc.sigma_pm) - 2*(2*pc.m2sq**2*pc.Q2 + pc.s1hat**2*(pc.Q2 + pc.s1hat) + pc.m2sq*pc.s1hat*(4*pc.Q2 + pc.s1hat))*pc.sigma_pp) + pc.s1hat*(pc.s1hat*(28*pc.m1sq**2*pc.Q2*pc.s1hat + pc.s1hat**2*(4*pc.m2sq**2 + 4*pc.m2sq*pc.s1hat + pc.Q2*pc.s1hat) + 2*pc.m1sq*(-12*pc.m2sq**2*pc.Q2 + pc.s1hat*(14*pc.Q2**2 - 9*pc.Q2*pc.s1hat + 2*pc.s1hat*(pc.s1hat - pc.sigma_pm)) + 2*pc.m2sq*pc.s1hat*(-16*pc.Q2 + pc.s1hat - pc.sigma_pm)) + (pc.s1hat*(20*pc.m2sq**2 + 20*pc.m2sq*pc.s1hat - 3*pc.Q2*pc.s1hat) + 4*pc.m2sq*pc.Q2*pc.sigma_mp)*pc.sigma_pm + 18*pc.m2sq*(pc.m2sq + pc.s1hat)*pc.sigma_pm**2) + (4*pc.m1sq*pc.m2sq*pc.Q2*pc.s1hat + pc.s1hat**2*(-pc.deltap**2 + 6*pc.Q2*(pc.m2sq + pc.s1hat) + 7*pc.Q2*pc.sigma_mp) - 8*pc.m2sq*pc.Q2*(2*pc.m2sq + 5*pc.s1hat)*pc.sigma_pm)*pc.sigma_pp + pc.s1hat*(-pc.deltap**2 + 6*pc.Q2*(pc.m2sq + pc.s1hat) - 3*pc.Q2*pc.sigma_mp)*pc.sigma_pp**2))*pc.x)/(2.*pc.deltap**2*pc.Q2*pc.s1hat*(pc.m2sq + pc.s1hat)**2) + (pc.L_xi*(-2*pc.delta**4*(pc.s1hat + pc.sigma_pp) - 2*pc.delta**2*(pc.s1hat + pc.sigma_pp)*(pc.s1hat**2 + 2*pc.s1hat*pc.sigma_pm - pc.Q2*pc.sigma_pp) + pc.s1hat*(pc.Q2*pc.s1hat**2*(pc.s1hat + 4*pc.sigma_pm) - pc.s1hat*(pc.deltap**2 - 7*pc.Q2*pc.sigma_pm)*pc.sigma_pp + 4*pc.Q2*pc.sigma_pm*pc.sigma_pp**2 + 2*pc.m1sq*pc.Q2*pc.s1hat*(-2*pc.m2sq + 6*pc.s1hat + 9*pc.sigma_pp)))*pc.x)/(pc.deltap**3*pc.Q2*pc.s1hat)

def fl_splus_soft(pc):
    return (-2*(2*pc.delta + pc.L_xisoft*pc.sigma_pp)*(pc.delta**2 - pc.Q2*pc.sigma_pp)*pc.x)/(pc.delta**2*pc.Q2)

def fl_splus_virt(pc):
    return ((pc.delta**2*(2*pc.CRm + pc.C1p*pc.m1sq + pc.C1m*pc.m2sq) + 4*pc.Cplus*np.sqrt(pc.m1sq * pc.m2sq)*pc.Q2 - 2*pc.CRm*pc.Q2*pc.sigma_pp)*pc.x)/(2.*pc.delta*pc.Q2)

def fl_sminus_raw(pc):
    return (-2*pc.L_xi*np.sqrt(pc.m1sq * pc.m2sq)*(pc.deltap**2*(2*pc.Q2 - pc.s1hat)*pc.s1hat + 2*pc.Q2*((3*pc.m1sq + pc.m2sq + pc.Q2)*pc.s1hat**2 + (pc.delta**2 + 2*pc.s1hat*pc.sigma_pm)*pc.sigma_pp))*pc.x)/(pc.deltap**3*pc.Q2*pc.s1hat) - (np.sqrt(pc.m1sq * pc.m2sq)*(4*pc.delta**2*pc.Q2*(pc.m2sq + pc.s1hat)*(2*pc.m2sq + pc.s1hat) + pc.s1hat*(pc.deltap**2*(4*pc.m2sq*(pc.Q2 - pc.s1hat) + pc.s1hat*(4*pc.Q2 - 3*pc.s1hat + pc.sigma_pp)) + 4*pc.Q2*(pc.m2sq + pc.s1hat)*(4*pc.m2sq*pc.sigma_pm + pc.s1hat*(2*pc.s1hat - pc.sigma_mp + 2*(pc.sigma_pm + pc.sigma_pp)))))*pc.x)/(pc.deltap**2*pc.Q2*pc.s1hat*(pc.m2sq + pc.s1hat)**2)

def fl_sminus_soft(pc):
    return (-4*np.sqrt(pc.m1sq * pc.m2sq)*(2*pc.delta + pc.L_xisoft*pc.sigma_pp)*pc.x)/pc.delta**2

def fl_sminus_virt(pc):
    return ((np.sqrt(pc.m1sq * pc.m2sq)*((pc.C1m + pc.C1p)*pc.delta**2 + 4*pc.CRm*pc.Q2) + 2*pc.Cplus*(pc.delta**2 - pc.Q2*pc.sigma_pp))*pc.x)/(2.*pc.delta*pc.Q2)

def f3_rplus_raw(pc):
    return (pc.s1hat*((-2*pc.delta**2*(2*pc.m2sq + pc.s1hat))/pc.s1hat**2 + pc.sigma_mp - 3*pc.sigma_pm - (2*(pc.delta**2 + 2*pc.m2sq*pc.sigma_pm))/pc.s1hat - ((pc.s1hat - pc.sigma_mp)*(pc.s1hat + pc.sigma_pp))/(2.*(pc.m2sq + pc.s1hat)))*pc.x)/(pc.deltap*(pc.m2sq + pc.s1hat)) - (pc.L_xi*(pc.s1hat*(pc.s1hat**2 - 4*pc.m1sq*pc.sigma_mp + 3*pc.s1hat*pc.sigma_pm) + 2*pc.delta**2*(2*pc.s1hat + pc.sigma_pp))*pc.x)/(pc.deltap**2*pc.s1hat)

def f3_rplus_soft(pc):
    return (-2*(2*pc.delta + pc.L_xisoft*pc.sigma_pp)*pc.x)/pc.delta

def f3_rplus_virt(pc):
    return pc.CRm*pc.x

def f3_rminus_raw(pc):
    return (2*np.sqrt(pc.m1sq * pc.m2sq)*(pc.s1hat - pc.sigma_mp)*pc.x)/(pc.deltap*(pc.m2sq + pc.s1hat)) + (2*pc.L_xi*np.sqrt(pc.m1sq * pc.m2sq)*(pc.s1hat + pc.sigma_pm)*pc.x)/pc.deltap**2

def f3_rminus_soft(pc):
    return 0

def f3_rminus_virt(pc):
    return pc.Cplus*pc.x

def m1_splus(pc):
    return pc.sigma_pp/(2.*pc.delta)

def m1_sminus(pc):
    return -((np.sqrt(pc.m1sq * pc.m2sq))/pc.delta)

def m2_splus(pc):
    return (pc.delta*pc.x)/pc.Q2

def m2_sminus(pc):
    return 0

def ml_splus(pc):
    return (pc.delta*pc.x)/pc.Q2 - (pc.sigma_pp*pc.x)/pc.delta

def ml_sminus(pc):
    return (2*np.sqrt(pc.m1sq * pc.m2sq)*pc.x)/pc.delta

def m3_rplus(pc):
    return pc.x

def m3_rminus(pc):
    return 0

def I1(pc):
    return np.log( ( pc.sigma_pp + pc.delta ) / ( pc.sigma_pp - pc.delta ) ) / pc.delta

def CRm(pc):
    return ( pc.delta**2 / 2 / pc.Q2 + pc.sigma_pp * ( 1 + np.log( pc.Q2 / pc.delta ) ) ) * pc.I1 + ( pc.m2sq - pc.m1sq ) / 2 / pc.Q2 * np.log( pc.m1sq / pc.m2sq ) - np.log( pc.Q2 / pc.m1sq ) - np.log( pc.Q2 / pc.m2sq ) - 4 + pc.sigma_pp / pc.delta * ( + np.log( np.abs( ( pc.delta - pc.sigma_pm ) / 2 / pc.Q2 ) )**2 / 2 + np.log( np.abs( ( pc.delta - pc.sigma_mp ) / 2 / pc.Q2 ) )**2 / 2 - np.log( np.abs( ( pc.delta + pc.sigma_pm ) / 2 / pc.Q2 ) )**2 / 2 - np.log( np.abs( ( pc.delta + pc.sigma_mp ) / 2 / pc.Q2 ) )**2 / 2 - li2( ( pc.delta - pc.sigma_pm ) / 2 / pc.delta ) - li2( ( pc.delta - pc.sigma_mp ) / 2 / pc.delta ) + li2( ( pc.delta + pc.sigma_pm ) / 2 / pc.delta ) + li2( ( pc.delta + pc.sigma_mp ) / 2 / pc.delta ) )

def Cplus(pc):
    return 2 * np.sqrt(pc.m1sq * pc.m2sq) * pc.I1

def C1m(pc):
    return - ( pc.sigma_pm * pc.I1 + np.log( pc.m1sq / pc.m2sq ) ) / pc.Q2

def C1p(pc):
    return - ( pc.sigma_mp * pc.I1 - np.log( pc.m1sq / pc.m2sq ) ) / pc.Q2

def S(pc):
    return 2 + pc.sigma_pp / pc.delta * ( pc.delta * pc.I1 + li2( 2 * pc.delta / ( pc.delta - pc.sigma_pp ) ) - li2( 2 * pc.delta / ( pc.delta + pc.sigma_pp ) ) ) + np.log( pc.delta**2 / pc.m2sq / pc.Q2 ) * ( - 2 + pc.sigma_pp * pc.I1 )



"""The asymptotic components for the FONLL prescription."""

from eko import basis_rotation as br

from .. import heavy, kernels, light


def import_pc_module(kind, process, subpkg=None):
    """Dynamic import implementing module, selected observable."""
    if subpkg is None:
        subpkg = __name__
    else:
        subpkg = ".".join(__name__.split(".")[:-2] + [subpkg, ""])
    return kernels.import_local(kind, process, subpkg)


def generate_missing_asy(esf, nf, ihq, pto_evol, icoupl=None):
    r"""
    Collect the high-virtuality limit of missing.

    Parameters
    ----------
    esf : EvaluatedStructureFunction
        kinematic point
    nf : int
        number of light flavors
    ihq : int
        heavy quark
    pto_evol : int
        PTO of evolution
    icoupl : None or int
        PID of the flavor coupling (default: None)

    Returns
    -------
    elems : list(yadism.kernels.Kernel)
        list of elements
    """
    # in CC there are no missing diagrams known yet
    if esf.process == "CC":
        return ()
    # only NC
    weights = light.kernels.nc_weights(
        esf.info.coupling_constants,
        esf.Q2,
        nf,
        esf.info.obs_name.is_parity_violating,
        skip_heavylight=True,
    )
    if icoupl is not None:
        weights["ns"] = {k: v for k, v in weights["ns"].items() if abs(k) == icoupl}

    kind = esf.info.obs_name.kind
    asy_cfs = import_pc_module(kind, esf.process)
    m2hq = esf.info.m2hq[ihq - 4]
    asys = []
    for res in range(pto_evol + 1):
        name = "Asy" + ("N" * res) + "LL" + "NonSinglet"
        km = kernels.Kernel(
            weights["ns"], asy_cfs.__getattribute__(name)(esf, nf, m2hq=m2hq)
        )
        asys.append(km)
    return asys


def generate_heavy_asy(esf, nf, pto_evol, ihq):
    """
     Collect the high-virtuality limit of heavy.

    Parameters
    ----------
        esf : EvaluatedStructureFunction
            kinematic point
        nf : int
            number of light flavors
        pto_evol : int
            PTO of evolution

    Returns
    -------
        elems : list(yadism.kernels.Kernel)
            list of elements
    """
    kind = esf.info.obs_name.kind
    is_pv = esf.info.obs_name.is_parity_violating
    asy_cfs = import_pc_module(kind, esf.process)
    m2hq = esf.info.m2hq[ihq - 4]
    asys = []
    if esf.process == "CC":
        wa = kernels.cc_weights(
            esf.info.coupling_constants,
            esf.Q2,
            br.quark_names[ihq - 1],
            nf,
            is_pv,
        )
        asys = [
            kernels.Kernel(wa["ns"], asy_cfs.AsyQuark(esf, nf, m2hq=m2hq)),
            kernels.Kernel(wa["g"], asy_cfs.AsyGluon(esf, nf, m2hq=m2hq)),
        ]
    else:
        asy_weights = heavy.kernels.nc_weights(
            esf.info.coupling_constants,
            esf.Q2,
            nf,
            ihq,
            is_pv,
        )
        if not is_pv:
            n3lo_cf_variation = esf.info.theory["n3lo_cf_variation"]
            for c, channel in (("g", "Gluon"), ("s", "Singlet")):
                for res in range(pto_evol + 1):
                    name = "Asy" + ("N" * res) + "LL" + channel
                    for av in ("AA", "VV"):
                        asys.append(
                            kernels.Kernel(
                                asy_weights[f"{c}{av}"],
                                asy_cfs.__getattribute__(name)(
                                    esf,
                                    nf,
                                    m2hq=m2hq,
                                    n3lo_cf_variation=n3lo_cf_variation,
                                ),
                            )
                        )
    return asys


def generate_intrinsic_asy(esf, nf, pto_evol, ihq):
    """
     Collect the high-virtuality limit of intrinsic, see also  :eqref:`10` of :cite:`nnpdf-intrinsic`.

    Parameters
    ----------
        esf : EvaluatedStructureFunction
            kinematic point
        nf : int
            number of light flavors
        pto_evol : int
            PTO of evolution

    Returns
    -------
        elems : list(yadism.kernels.Kernel)
            list of elements
    """
    kind = esf.info.obs_name.kind
    is_pv = esf.info.obs_name.is_parity_violating
    cfs = import_pc_module(kind, esf.process)
    m2hq = esf.info.m2hq[ihq - 4]
    if esf.process == "CC":
        w = kernels.cc_weights(
            esf.info.coupling_constants,
            esf.Q2,
            br.quark_names[ihq - 1],
            ihq,
            is_pv,
        )
        weights = {k: v for k, v in w["ns"].items() if abs(k) == ihq}
    else:  # NC
        if is_pv:
            wVA = esf.info.coupling_constants.get_weight(ihq, esf.Q2, "VA")
            wAV = esf.info.coupling_constants.get_weight(ihq, esf.Q2, "AV")
            wp = wVA + wAV
            # Rminus ~ wm = wVA - wAV does not contribute to asy
            weights = {ihq: wp, (-ihq): -wp}
        else:
            wVV = esf.info.coupling_constants.get_weight(ihq, esf.Q2, "VV")
            wAA = esf.info.coupling_constants.get_weight(ihq, esf.Q2, "AA")
            wp = wVV + wAA
            # Sminus ~ wm = wVV - wAA does not contribute to asy
            weights = {ihq: wp, (-ihq): wp}

    asys = [
        kernels.Kernel(
            weights,
            cfs.AsyLLIntrinsic(esf, nf, m2hq=m2hq),
        )
    ]
    if pto_evol > 0:
        asys.extend(
            [
                kernels.Kernel(
                    weights,
                    cfs.AsyNLLIntrinsicMatching(esf, nf, m2hq=m2hq),
                ),
                kernels.Kernel(
                    weights,
                    cfs.AsyNLLIntrinsicLight(esf, nf, m2hq=m2hq),
                ),
            ]
        )
    return asys



"""Export predictions to PineAPPL."""

import json

import numpy as np

import yadism
from yadism import observable_name


def dump_pineappl_to_file(output, filename, obsname):
    """Write output on a PineAPPL grid file.

    Parameters
    ----------
    filename : str
        output file name
    obsname : str
        observable to be dumped

    """
    # pylint: disable=no-member, too-many-locals
    if len(output[obsname]) <= 0:
        raise ValueError(f"no ESF {obsname}!")
    import pineappl  # pylint: disable=import-outside-toplevel,import-error

    interpolation_xgrid = output["xgrid"]["grid"]
    # interpolation_is_log = self["interpolation_is_log"]
    interpolation_polynomial_degree = output["polynomial_degree"]

    # Instantiate the objects required to construct a new Grid
    channels = [pineappl.boc.Channel([([pid], 1.0)]) for pid in output["pids"]]
    first_esf_result = output[obsname][0]
    orders = [pineappl.boc.Order(*o, 0) for o in first_esf_result.orders]
    polarized = obsname.startswith("g")
    convolution_types = pineappl.convolutions.ConvType(
        polarized=polarized, time_like=False
    )
    convolutions = [
        pineappl.convolutions.Conv(convolution_types=convolution_types, pid=2212)
    ]
    kinematics = [pineappl.boc.Kinematics.Scale(0), pineappl.boc.Kinematics.X(0)]
    scale_funcs = pineappl.boc.Scales(
        ren=pineappl.boc.ScaleFuncForm.Scale(0),
        fac=pineappl.boc.ScaleFuncForm.Scale(0),
        frg=pineappl.boc.ScaleFuncForm.NoScale(0),
    )
    bins = len(output[obsname])
    bin_limits = pineappl.boc.BinsWithFillLimits.from_fill_limits(
        fill_limits=list(map(float, range(0, bins + 1)))
    )

    q2grid = np.array([obs.Q2 for obs in output[obsname]])
    interpolations = [
        pineappl.interpolation.Interp(
            min=q2grid.min(),
            max=q2grid.max(),
            nodes=50,
            order=3,
            reweight_meth=pineappl.interpolation.ReweightingMethod.NoReweight,
            map=pineappl.interpolation.MappingMethod.ApplGridH0,
            interpolation_meth=pineappl.interpolation.InterpolationMethod.Lagrange,
        ),  # Interpolation on the Scale
        pineappl.interpolation.Interp(
            min=interpolation_xgrid[0],
            max=interpolation_xgrid[-1],
            nodes=len(interpolation_xgrid),
            order=interpolation_polynomial_degree,
            reweight_meth=pineappl.interpolation.ReweightingMethod.ApplGridX,
            map=pineappl.interpolation.MappingMethod.ApplGridF2,
            interpolation_meth=pineappl.interpolation.InterpolationMethod.Lagrange,
        ),  # Interpolation on momentum fraction x
    ]

    grid = pineappl.grid.Grid(
        pid_basis=pineappl.pids.PidBasis.Pdg,
        channels=channels,
        orders=orders,
        bins=bin_limits,
        convolutions=convolutions,
        interpolations=interpolations,
        kinematics=kinematics,
        scale_funcs=scale_funcs,
    )
    limits = []
    on = observable_name.ObservableName(obsname)
    is_xs = on.kind in observable_name.xs

    # add each ESF as a bin
    for bin_, obs in enumerate(output[obsname]):
        x = obs.x
        Q2 = obs.Q2
        kins = [(Q2, Q2), (x, x)]
        if is_xs:
            kins.append((obs.y, obs.y))
        limits.append(kins)

        # add all orders
        for o, (v, _e) in obs.orders.items():
            order_index = list(first_esf_result.orders.keys()).index(o)
            prefactor = (
                ((1.0 / (4.0 * np.pi)) ** o[0]) * ((-1.0) ** o[2]) * ((-1.0) ** o[3])
            )
            # add for each pid/lumi
            for pid_index, pid_values in enumerate(v):
                pid_values = prefactor * pid_values
                # grid is empty? skip
                if not any(np.array(pid_values) != 0):
                    continue
                subgrid = pineappl.subgrid.ImportSubgridV1(
                    array=pid_values[np.newaxis, :],
                    node_values=[[Q2], interpolation_xgrid],
                )
                grid.set_subgrid(order_index, bin_, pid_index, subgrid.into())
    # set the correct observables
    normalizations = [1.0] * bins
    bin_configs = pineappl.boc.BinsWithFillLimits.from_limits_and_normalizations(
        limits=limits,
        normalizations=normalizations,
    )
    grid.set_bwfl(bin_configs)

    # set metadata
    grid.set_metadata("theory", json.dumps(output.theory))
    grid.set_metadata("runcard", json.dumps(output.observables))
    grid.set_metadata("yadism_version", yadism.__version__)
    # set bin information
    grid.set_metadata("x1_label", "Q2")
    grid.set_metadata("x1_label_tex", "$Q^2$")
    grid.set_metadata("x1_unit", "GeV^2")
    grid.set_metadata("x2_label", "x")
    grid.set_metadata("x2_label_tex", "$x$")
    grid.set_metadata("x2_unit", "")
    if is_xs:
        grid.set_metadata("x3_label", "y")
        grid.set_metadata("x3_label_tex", "$y$")
        grid.set_metadata("x3_unit", "")
    grid.set_metadata("y_label", obsname)

    # dump file
    grid.optimize()
    grid.write_lz4(filename)



from ..light import f2_nc as light
from ..partonic_channel import RSL, EmptyPartonicChannel
from . import g1_nc_raw as raw
from . import partonic_channel as pc


class AsyLLNonSinglet(pc.PartonicChannelAsy):
    def NNLO(self):
        return RSL(raw.c2ns_LL_reg, raw.c2ns_LL_sing, raw.c2ns_LL_loc, args=[self.L])


class AsyNLLNonSinglet(pc.PartonicChannelAsy):
    def NNLO(self):
        return RSL(raw.c2ns_NLL_reg, raw.c2ns_NLL_sing, raw.c2ns_NLL_loc, args=[self.L])


class AsyNNLLNonSinglet(pc.PartonicChannelAsy):
    def NNLO(self):
        return RSL(
            raw.c2ns_NNLL_reg, raw.c2ns_NNLL_sing, raw.c2ns_NNLL_loc, args=[self.L]
        )


class AsyLLSinglet(pc.PartonicChannelAsy):
    def NNLO(self):
        return RSL(raw.c2ps_LL_reg, args=[self.L])


class AsyNLLSinglet(pc.PartonicChannelAsy):
    def NNLO(self):
        return RSL(raw.c2ps_NLL_reg, args=[self.L])


class AsyNNLLSinglet(pc.PartonicChannelAsy):
    def NNLO(self):
        return RSL(raw.c2ps_NNLL_reg, args=[self.L])


class AsyLLGluon(pc.PartonicChannelAsy):
    def NLO(self):
        return RSL(raw.c1g_LL_reg, args=[self.L])

    def NNLO(self):
        return RSL(raw.c2g_LL_reg, args=[self.L])


class AsyNLLGluon(pc.PartonicChannelAsy):
    def NLO(self):
        return RSL(raw.c1g_NLL_reg, args=[self.L])

    def NNLO(self):
        return RSL(raw.c2g_NLL_reg, args=[self.L])


class AsyNNLLGluon(pc.PartonicChannelAsy):
    def NNLO(self):
        return RSL(raw.c2g_NNLL_reg, args=[self.L])


class AsyLLIntrinsic(EmptyPartonicChannel):
    pass


class AsyNLLIntrinsicMatching(EmptyPartonicChannel):
    pass


class AsyNLLIntrinsicLight(EmptyPartonicChannel):
    pass



"""Partonic channels of the massless limit."""

import numpy as np
from eko import constants

from .. import partonic_channel as pc
from .. import splitting_functions as split
from ..partonic_channel import RSL


class PartonicChannelAsy(pc.PartonicChannel):
    """Massless limit of a coeficient function."""

    def __init__(self, *args, m2hq, n3lo_cf_variation=0):
        super().__init__(*args)
        self.L = np.log(self.ESF.Q2 / m2hq)
        self.n3lo_cf_variation = n3lo_cf_variation


class PartonicChannelAsyLLIntrinsic(PartonicChannelAsy):
    """|ref| implements |LL| part of :eqref:`10` of :cite:`nnpdf-intrinsic` from matching."""

    light_cls = lambda _esf, _nf: None

    def LO(self):
        """Return |LO| from light."""
        return self.light_cls(self.ESF, self.nf).LO()

    def lo_local(self):
        """Collect LO delta coefficient."""
        # don't rely on self.LO() since is overwritten in NLL
        lo = self.light_cls(self.ESF, self.nf).LO()
        # check LO is not vanishing
        if lo is None:
            return None
        # now, then it can only be a delta - and the delta is coming from the arguments
        return lo.args["loc"][0]

    def NLO(self):
        """|ref| implements :eqref:`10` of :cite:`nnpdf-intrinsic`."""
        lo_local = self.lo_local()
        if lo_local is None:
            return None
        l = self.L

        # since LO is a delta it is sufficient to multiply with its coefficient
        def reg(z, args):
            return lo_local * l * split.lo.pqq_reg(z, args)

        def sing(z, args):
            return lo_local * l * split.lo.pqq_sing(z, args)

        def loc(x, args):
            return lo_local * l * split.lo.pqq_local(x, args)

        return RSL(reg, sing, loc)

    @staticmethod
    def NNLO():
        """Empty, because intrinsic is only known up to |NLO|."""
        return None

    @staticmethod
    def N3LO():
        """Empty, because intrinsic is only known up to |NLO|."""
        return None


class PartonicChannelAsyNLLIntrinsicMatching(PartonicChannelAsyLLIntrinsic):
    """|ref| implements |NLL| part of :eqref:`10` of :cite:`nnpdf-intrinsic` from matching."""

    @staticmethod
    def LO():
        """Empty, because |NLL| only starts at |NLO|."""
        return None

    def NLO(self):
        """|ref| implements |NLL| part of :eqref:`10` of :cite:`nnpdf-intrinsic` from matching."""
        lo_local = self.lo_local()
        if lo_local is None:
            return None
        asnorm = 2.0

        # again, convolution with LO is trivial as it is a delta
        def sing(z, _args):
            # this coefficient function is proportional to P_qq
            # i.e. 2CF * (1.0 + z ** 2) / (1.0 - z) is the "bare" P_qq
            return (
                lo_local
                * asnorm
                * constants.CF
                * ((1.0 + z**2) / (1.0 - z) * (-2.0 * np.log(1.0 - z) - 1.0))
            )

        # MMa:
        # FortranForm@FullSimplify@Integrate[(1 + z^2)/(1 - z) (- 2 Log[1 - z] - 1), {z, 0, x}, Assumptions -> {0 < x < 1}] # pylint: disable=line-too-long
        def loc(x, _args):
            return (
                -1.0
                * lo_local
                * asnorm
                * constants.CF
                * (
                    -x * 2.0
                    + np.log(1.0 - x) * (-1.0 + x * (2.0 + x) + 2.0 * np.log(1.0 - x))
                )
            )

        return RSL(sing=sing, loc=loc)


class PartonicChannelAsyNLLIntrinsicLight(PartonicChannelAsyLLIntrinsic):
    """|ref| implements |NLL| part of :eqref:`10` of :cite:`nnpdf-intrinsic` from light.

    In practice it echos the |NLO| light coefficient function only.
    """

    @staticmethod
    def LO():
        """Empty, because |NLL| only starts at |NLO|."""
        return None

    def NLO(self):
        """|ref| implements |NLL| part of :eqref:`10` of :cite:`nnpdf-intrinsic` from light."""
        return self.light_cls(self.ESF, self.nf).NLO()


class NeutralCurrentBaseAsy(PartonicChannelAsy):
    pass



r"""This module contains the main loop for the DIS calculations.

There are two ways of using ``yadism``:

* ``Runner``: this class provides a runner that get the *theory* and
  *observables* descriptions as input and manage the whole observables'
  calculation process
* ``run_dis``: a function that wraps the construction of a ``Runner`` object
  and calls the proper method to get the requested output

.. todo::
    decide about ``run_dis`` and document it properly in module header
"""

import copy
import inspect
import io
import logging
import time

import numpy as np
import rich
import rich.align
import rich.box
import rich.console
import rich.markdown
import rich.panel
import rich.progress
from eko import basis_rotation as br
from eko import matchings
from eko.interpolation import InterpolatorDispatcher, XGrid
from eko.quantities.heavy_quarks import MatchingScales

from . import log, observable_name
from .coefficient_functions.coupling_constants import CouplingConstants
from .esf import scale_variations as sv
from .input import compatibility
from .output import Output
from .sf import StructureFunction as SF
from .xs import CrossSection as XS

logger = logging.getLogger(__name__)


class Runner:
    """Wrapper to compute a process.

    Parameters
    ----------
    theory : dict
        Dictionary with the theory parameters for the evolution (currently
        including PDFSet and DIS process indication).
    observables : dict
        DIS parameters: process description, kinematic specification for the
        requested output.

    Notes
    -----
    For a full description of the content of `theory` and `dis_observables`
    dictionaries read ??.

    .. todo::
        * reference on theory template
        * detailed description of dis_observables entries
    """

    banner = rich.align.Align(
        rich.panel.Panel.fit(
            inspect.cleandoc(
                r"""  __     __       _ _
                     \ \   / /      | (_)
                      \ \_/ /_ _  __| |_ ___ _ __ ___
                       \   / _` |/ _` | / __| '_ ` _ \
                        | | (_| | (_| | \__ \ | | | | |
                        |_|\__,_|\__,_|_|___/_| |_| |_|
                """
            ),
            rich.box.SQUARE,
            padding=1,
            style="magenta",
        ),
        "center",
    )

    def __init__(self, theory: dict, observables: dict):
        new_theory, new_observables = compatibility.update(theory, observables)

        # Store inputs
        self._theory = new_theory
        self._observables = new_observables

        # Setup eko stuffs
        xgrid = XGrid(
            self._observables["interpolation_xgrid"],
            self._observables["interpolation_is_log"],
        )
        interpolator = InterpolatorDispatcher(
            xgrid, self._observables["interpolation_polynomial_degree"], mode_N=False
        )

        # Non-eko theory
        coupling_constants = CouplingConstants.from_dict(theory, self._observables)
        pto = new_theory["PTODIS"]
        pto_evol = new_theory["PTO"]
        sv_manager = sv.ScaleVariations(
            order=pto,
            interpolator=interpolator,
            activate_ren=new_theory["RenScaleVar"],
            activate_fact=new_theory["FactScaleVar"],
        )

        # Initialize structure functions
        masses = np.power([new_theory["mc"], new_theory["mb"], new_theory["mt"]], 2)
        thresholds_ratios = np.power(
            [new_theory["kcThr"], new_theory["kbThr"], new_theory["ktThr"]], 2
        )
        managers = dict(
            interpolator=interpolator,
            threshold=matchings.Atlas(
                matching_scales=MatchingScales(list(masses * thresholds_ratios)),
                origin=(new_theory["Q0"] ** 2, new_theory["nf0"]),
            ),
            coupling_constants=coupling_constants,
            sv_manager=sv_manager,
        )
        # pass theory params
        theory_params = dict(
            pto=pto,
            pto_evol=pto_evol,
            scheme=theory["FNS"],
            nf_ff=theory["NfFF"],
            ZMq=(new_theory["ZMc"], new_theory["ZMb"], new_theory["ZMt"]),
            m2hq=masses,
            TMC=theory["TMC"],
            target=new_observables["TargetDIS"],
            GF=theory["GF"],
            M2W=theory["MW"] ** 2,
            M2target=theory["MP"] ** 2,
            fonllparts=new_theory["FONLLParts"],
            n3lo_cf_variation=theory["n3lo_cf_variation"],
        )
        logger.info(
            "PTO: %d, PTO@evolution: %d, process: %s",
            pto,
            pto_evol,
            new_observables["prDIS"],
        )
        self.configs = RunnerConfigs(theory=theory_params, managers=managers)
        logger.info("FNS: %s, NfFF: %d", theory["FNS"], theory["NfFF"])
        logger.info(
            "projectile: %s, target: {Z: %g, A: %g}",
            new_observables["ProjectileDIS"],
            *new_observables["TargetDIS"].values(),
        )

        self.observables = {}
        for obs_name, kins in self._observables["observables"].items():
            on = observable_name.ObservableName(obs_name)
            if on.kind in observable_name.xs:
                obs = XS(on, self)
            else:
                # TODO use get_sf?
                obs = SF(on, self)
            # read kinematics
            obs.load(kins)
            self.observables[obs_name] = obs

        # output console
        if log.silent_mode:
            file = io.StringIO()
        else:
            file = None
        self.console = rich.console.Console(file=file)
        log.setup(self.console)
        # ==============================
        # Initialize output
        # ==============================
        self._output = Output()
        self._output.theory = theory
        self._output.observables = observables
        self._output.update(interpolator.to_dict())
        self._output["pids"] = br.flavor_basis_pids
        self._output["projectilePID"] = coupling_constants.obs_config["projectilePID"]

    def get_sf(self, obs_name):
        """Return associated SF object."""
        if obs_name.name not in self.observables:
            self.observables[obs_name.name] = SF(obs_name, self)
        return self.observables[obs_name.name]

    def drop_cache(self):
        """Drop the whole cache for all observables.

        This preserves final results, since they are not part of the
        cache.
        """
        for obs in self.observables.values():
            if isinstance(obs, SF):
                obs.drop_cache()

    def replace_nans_with_0(self, out):
        """Replace any NaNs in output with 0.0

        The small-x (i.e. large eta) limit is not addressed in LeProHQ because
        the high energy limit of the polarized case is not known and that is the
        main purpose of LeProHQ. As a result LeProHQ may return NaNs in the
        small-x limit, which this function replaces with 0.0.

        Note that the value of x where this plays a role is below the
        experimental regime and thus this does not affect the description of
        data, but only e.g. the grids used for the FIATLUX photon computation.
        """

        out2 = copy.deepcopy(out)

        # Loop through each observable in the dictionary
        for observable, points in out2.items():
            # Skip the keys that are not an observable
            if observable not in observable_name.kinds:
                continue

            # Loop over the kinematic points
            for point in points:
                # Loop over each perturbative order
                for values in point.orders.values():
                    # `values` is a tuple containing the computed values and the integration error
                    for tup in range(2):
                        # Set any NaN or inf values in the array to 0
                        values[tup][~np.isfinite(values[tup])] = 0.0
        return out2

    def get_result(self):
        """Compute coefficient functions grid for requested kinematic points.

        Returns
        -------
        :obj:`Output`
            output object, it will store the coefficient functions grid
            (flavour, interpolation-index) for each requested kinematic
            point (x, Q2)
        """
        self.console.print(self.banner)

        # precomputing the plan of calculation
        precomputed_plan = {}
        printable_plan = []
        for name, obs in self.observables.items():
            if name in self._observables["observables"].keys():
                precomputed_plan[name] = obs
                printable_plan.append(
                    f"- {name} at {len(self._observables['observables'][name])} pts"
                )

        self.console.print(rich.markdown.Markdown("## Plan"))
        self.console.print(rich.markdown.Markdown("\n".join(printable_plan)))

        # running the calculation
        self.console.print(rich.markdown.Markdown("## Calculation"))
        self.console.print("yadism took off! please stay tuned ...")
        start = time.time()

        with rich.progress.Progress(transient=True, console=self.console) as progress:
            task = progress.add_task(
                "Starting...",
                total=sum(len(obs) for obs in precomputed_plan.values()),
            )

            for name, obs in precomputed_plan.items():
                # init slots
                results = [None] * len(obs)
                Q2 = None

                # compute
                for idx, elem in sorted(
                    enumerate(obs.elements), key=lambda indexed: indexed[1].Q2
                ):
                    # if we're changing Q2, drop cache
                    if Q2 is not None and Q2 != elem.Q2:
                        self.drop_cache()
                    Q2 = elem.Q2

                    results[idx] = elem.get_result()
                    progress.update(
                        task,
                        description=f"Computing [bold green]{name}",
                        advance=1,
                    )

                self.drop_cache()
                self._output[name] = results

        end = time.time()
        diff = end - start
        self.console.print(f"[cyan]took {diff:.2f} s")

        out = copy.deepcopy(self._output)
        out = self.replace_nans_with_0(out)
        return out


class RunnerConfigs:
    """Runner Configuration."""

    def __init__(self, theory, managers):
        self.theory = theory
        self.managers = managers

    def __getattribute__(self, name):
        managers = object.__getattribute__(self, "managers")
        if name == "managers":
            return managers
        if name in managers:
            return self.managers[name]

        theory = object.__getattribute__(self, "theory")
        if name == "theory":
            return theory
        if name in theory:
            return theory[name]

        return super().__getattribute__(name)



# auto-generated module by light package
# pylint: skip-file
# fmt: off
import numba as nb
import numpy as np


@nb.njit("f8(f8,f8[:])", cache=True)
def clnn2a(y, args):
    nf = args[0]
    dl  = np.log(y)
    dl1 = np.log(1.-y)
    res = - 40.41 + 97.48 * y + (26.56 * y - 0.031) * dl**2 - 14.85 * dl + 13.62 * dl1**2 - 55.79 * dl1 - 150.5 * dl * dl1 + nf * 16./27.e0 * ( 6.* y*dl1 - 12.* y*dl - 25.* y + 6.)
    return res

@nb.njit("f8(f8,f8[:])", cache=True)
def clnc2a(y, args):
    nf = args[0]
    dl  = np.log(y)
    dl1 = np.log(1.-y)
    res = - 52.27 + 100.8 * y + (23.29 * y - 0.043) * dl**2 - 22.21 * dl + 13.30 * dl1**2 - 59.12 * dl1 - 141.7 * dl * dl1 + nf * 16./27.e0 * ( 6.* y*dl1 - 12.* y*dl - 25.* y + 6.)
    return res

@nb.njit("f8(f8,f8[:])", cache=True)
def clnn2c(y, args):
    res = -0.164
    return res

@nb.njit("f8(f8,f8[:])", cache=True)
def clnc2c(y, args):
    res = -0.150
    return res



"""
Note that the source files (given in fortran) follow the notations in :cite:`vogt-f3cc`
where the odd-N moments are called minus even if they correspond to :math:`\nu + \bar{\nu}`,
while even-N moments are called plus even if they correspond to :math:`\nu - \bar{\nu}`.
This convention is changed in :cite:`Davies:2016ruz` where the |N3LO| CC results are presented
for the first time.
"""

from .. import partonic_channel as pc
from ..partonic_channel import RSL
from . import f3_nc, n3lo, nnlo


class NonSingletOdd(f3_nc.NonSinglet):
    pass


class NonSingletEven(f3_nc.NonSinglet):
    def NNLO(self):
        """|ref| implements the sum between :eqref:`2.8` and :eqref:`3.5`, :cite:`Davies:2016ruz`."""
        return RSL(
            nnlo.xc3ns2p.c3np2a, nnlo.xc3ns2p.c3ns2b, nnlo.xc3ns2p.c3np2c, [self.nf]
        )

    def N3LO(self):
        """|ref| implements the sum between :eqref:`2.11` and :eqref:`3.8`, :cite:`Davies:2016ruz`."""
        return RSL(
            n3lo.xc3ns3p.c3np3a,
            n3lo.xc3ns3p.c3ns3b,
            n3lo.xc3ns3p.c3np3c,
            [self.nf],
        )


class Gluon(pc.EmptyPartonicChannel):
    pass


class Singlet(pc.EmptyPartonicChannel):
    pass


class Valence(f3_nc.Valence):
    pass



from ..partonic_channel import RSL
from . import n3lo, nlo, nnlo
from . import partonic_channel as pc


class NonSinglet(pc.LightBase):
    @staticmethod
    def NLO():
        """
        |ref| implements :eqref:`3`, :cite:`vogt-flnc`.
        """

        return RSL(nlo.fl.ns_reg)

    def NNLO(self):
        """
        |ref| implements :eqref:`4`, :cite:`vogt-flnc`.
        """

        return RSL(
            nnlo.xclns2p.clnn2a, loc=nnlo.xclns2p.clnn2c, args=dict(reg=[self.nf])
        )

    def N3LO(self):
        """
        |ref| implements the flavor class :math:`fl_{2}` of :eqref:`8`, :cite:`vogt-flnc`.
        """
        return RSL(
            n3lo.xclns3p.clnp3a_fl2,
            loc=n3lo.xclns3p.clnp3c_fl2,
            args=dict(reg=[self.nf], loc=[self.nf]),
        )


class Gluon(pc.LightBase):
    def NLO(self):
        """
        |ref| implements :eqref:`3`, :cite:`vogt-flnc`.
        """

        return RSL(nlo.fl.gluon_reg, args=[self.nf])

    def NNLO(self):
        """
        |ref| implements :eqref:`6`, :cite:`vogt-flnc`.
        """

        return RSL(nnlo.xclsg2p.clg2a, args=[self.nf])

    def N3LO(self):
        """
        |ref| implements the flavor class :math:`fl_{2}` of :eqref:`10`, :cite:`vogt-flnc`.
        """
        return RSL(n3lo.xclsg3p.clg3a_fl2, args=[self.nf])


class Singlet(pc.LightBase):
    def NNLO(self):
        """
        |ref| implements :eqref:`5`, :cite:`vogt-flnc`.
        """

        return RSL(nnlo.xclsg2p.cls2a, args=[self.nf])

    def N3LO(self):
        """
        |ref| implements the flavor class :math:`fl_{2}` of :eqref:`9`, :cite:`vogt-flnc`.
        """
        return RSL(n3lo.xclsg3p.cls3a_fl2, args=[self.nf])


class GluonFL11(pc.LightBase):
    """Gluon flavor class :math:`fl_{11}`."""

    def N3LO(self):
        """
        |ref| implements the flavor class :math:`fl_{11}` of :eqref:`10`, :cite:`vogt-flnc`.
        """
        return RSL(n3lo.xclsg3p.clg3a_fl11, args=[self.nf])


class QuarkFL11(pc.LightBase):
    """Quark flavor class :math:`fl_{11}`."""

    def N3LO(self):
        """
        |ref| implements the flavor class :math:`fl_{11}` of :eqref:`9`, :cite:`vogt-flnc`.
        """
        return RSL(n3lo.xclsg3p.cls3a_fl11, args=[self.nf])



"""Benchmark to xspace_bench (the original FONLL implementation)."""

import numpy as np
from eko.couplings import Couplings, couplings_mod_ev
from eko.io import dictlike, runcards, types
from eko.matchings import Atlas, nf_default
from eko.quantities.heavy_quarks import MatchingScales

from yadism import observable_name as on


def compute_xspace_bench_data(theory, observables, pdf):
    """
    Run xspace_bench to compute observables.

    Parameters
    ----------
    theory : dict
        theory runcard
    observables : dict
        observables runcard
    pdf : lhapdf_like
        PDF set

    Returns
    -------
    dict
        xspace_bench numbers
    """
    import xspace_bench  # pylint:disable=import-outside-toplevel,import-error

    # Available targets: PROTON, NEUTRON, ISOSCALAR, IRON
    target = "PROTON"

    # Available projectiles: electron, neutrino, positron, antineutrino
    proj = observables["ProjectileDIS"].upper()

    # Available processes: EM, CC, NC
    proc = observables["prDIS"]

    # pto = 0,1 LO, NLO
    pto = theory["PTO"]
    if pto != 0 and pto != 1:
        raise NotImplementedError(
            f"{pto} not implemented in xspace_bench, use 0 for LO, 1 for NLO"
        )

    pdf_name = pdf.set().name
    if pdf_name == "ToyLH":
        raise Warning("yadmark ToyLH not equal to xspace_bench ToyLH")

    # Constant values
    q_thr = []
    q_thr.append(theory["mc"])
    q_thr.append(theory["mb"])
    q_thr.append(theory["mt"])
    mz = theory["MZ"]
    mw = theory["MW"]
    sw = theory["SIN2TW"]
    gf = theory["GF"]

    ckm = theory["CKM"].split(" ")
    ckm = [float(item) for item in ckm]

    # FONLL damping, otherwise set to 0
    damp = 0

    # select scheme
    scheme = theory["FNS"]

    if scheme == "ZM-VFNS":
        scheme = "ZMVN"

    elif scheme == "FFNS":
        if theory["NfFF"] != 3:
            raise NotImplementedError("FFNS only with 3 light flavors in xspace_bench")

    elif scheme == "FONLL-A":
        if theory["NfFF"] != 4:
            raise NotImplementedError(
                "FONLL-A only with 3 ( ie. NfFF=4 ) light flavors in xspace_bench"
            )
        damp = theory["DAMP"]
        scheme = "GMVN"
    else:
        raise NotImplementedError(f"{scheme} is not implemented in xspace_bench.")

    new_eko_theory = runcards.Legacy(theory=theory, operator={}).new_theory
    method = runcards.Legacy.MOD_EV2METHOD.get(theory["ModEv"], theory["ModEv"])
    method = dictlike.load_enum(types.EvolutionMethod, method)
    method = couplings_mod_ev(method)
    masses = [mq**2 for mq, _ in new_eko_theory.heavy.masses]
    thresholds_ratios = np.power(new_eko_theory.heavy.matching_ratios, 2)
    sc = Couplings(
        couplings=new_eko_theory.couplings,
        order=new_eko_theory.order,
        method=method,
        masses=masses,
        hqm_scheme=new_eko_theory.heavy.masses_scheme,
        thresholds_ratios=thresholds_ratios.tolist(),
    )
    atlas = Atlas(
        matching_scales=MatchingScales(masses * thresholds_ratios),
        origin=(theory["Qref"] ** 2, theory["nfref"]),
    )
    num_tab = {}
    # loop over functions
    for obs_name in observables["observables"]:
        # if not on.ObservableName.is_valid(obs):
        #    continue

        obs = on.ObservableName(obs_name)

        out = []
        q2s = []
        # get all the q2
        for kin in observables["observables"].get(obs_name, []):
            if kin["Q2"] not in q2s:
                q2s.append(kin["Q2"])

        # loop over points
        for q2 in q2s:
            xs = []

            alphas = sc.a_s(q2, nf_to=nf_default(q2, atlas)) * 4.0 * np.pi
            y = 0.5
            f = 0.0

            # get all the x corresponding to q2
            for kin in observables["observables"].get(obs_name, []):
                if kin["Q2"] == q2:
                    xs.append(kin["x"])

            for x in xs:
                res = []
                f3_fact = -1.0
                if x == 1.0:
                    res = np.zeros((3, 5))

                elif proc == "NC" or proc == "EM":
                    f3_fact = 1.0
                    res = xspace_bench.nc_dis(
                        x,
                        q2,
                        y,
                        proc,
                        scheme,
                        pto,
                        pdf_name,
                        target,
                        proj,
                        mz,
                        sw,
                        alphas,
                        q_thr,
                        damp,
                    )
                elif proc == "CC":
                    # for positron F3 has opposite sign
                    if proj == "POSITRON" or proj == "ANTINEUTRINO":
                        f3_fact = 1.0
                    res = xspace_bench.cc_dis(
                        x,
                        q2,
                        y,
                        scheme,
                        pto,
                        pdf_name,
                        target,
                        proj,
                        gf,
                        mw,
                        ckm,
                        alphas,
                        q_thr,
                        damp,
                    )
                else:
                    raise NotImplementedError(
                        f"{proc} is not implemented in xspace_bench "
                    )

                if obs.kind == "F2":
                    if obs.flavor == "light":
                        f = res[0][0]
                    if obs.flavor == "charm":
                        f = res[0][1]
                    if obs.flavor == "total":
                        f = res[0][4]
                elif obs.kind == "F3":
                    if obs.flavor == "light":
                        f = f3_fact * res[1][0]
                    if obs.flavor == "charm":
                        f = f3_fact * res[1][1]
                    if obs.flavor == "total":
                        f = f3_fact * res[1][4]
                elif obs.kind == "FL":
                    if obs.flavor == "light":
                        f = res[2][0]
                    if obs.flavor == "charm":
                        f = res[2][1]
                    if obs.flavor == "total":
                        f = res[2][4]
                else:
                    raise NotImplementedError(
                        f"{obs.kind} is not implemented in xspace_bench"
                    )
                out.append(dict(x=x, Q2=q2, result=f))

        num_tab[obs_name] = out

    return num_tab



import pathlib

import numpy as np
from scipy.interpolate import RectBivariateSpline

grid_path = pathlib.Path(__file__).parent / "grids"

eta_grid = np.load(grid_path / "eta.npy")
xi_grid = np.load(grid_path / "xi.npy")

interpolators = {}


def interpolator(coeff, nf, variation):
    grid_name = f"{coeff}_nf{int(nf)}_var{int(variation)}.npy"

    # is it already loaded ?
    if grid_name in interpolators:
        return interpolators[grid_name]

    # load grid
    coeff = np.load(grid_path / grid_name)
    grid_interpolator = RectBivariateSpline(xi_grid, eta_grid, coeff)

    # store result
    interpolators[grid_name] = grid_interpolator

    return grid_interpolator



"""Utilities to help run APFEL++ benchmarks."""

import numpy as np
from eko import basis_rotation as br

# Q2 knots specs
NQ = 250
QMIN = 1
QMAX = 200

# Map observables names to APFEL++ methods
MAP_HEAVINESS = {
    "charm": 4,
    "bottom": 5,
    "top": 6,
    "light": 0,  # TODO: Check combination
    "total": 0,
}

PROJECTILE_PIDS = {
    "electron": 11,
    "positron": -11,
    "neutrino": 12,
    "antineutrino": -12,
}


def map_apfelpy_sf(init, observables, fns):
    """Get required structure function.

    Parameters
    ----------
    init: ap.initializers
        Apfel++ initializers
    observables: dict
        observables runcard
    fns: str
        flavor number scheme

    Returns
    -------
        Apfel++ structure function initalizer
    """
    MAP_ZM_CC = {
        "F2m": init.InitializeF2CCMinusObjectsZM,
        "FLm": init.InitializeFLCCMinusObjectsZM,
        "F3m": init.InitializeF3CCMinusObjectsZM,
        "F2p": init.InitializeF2CCPlusObjectsZM,
        "FLp": init.InitializeFLCCPlusObjectsZM,
        "F3p": init.InitializeF3CCPlusObjectsZM,
    }

    MAP_ZM_NC = {
        "F2": init.InitializeF2NCObjectsZM,
        "FL": init.InitializeFLNCObjectsZM,
        "F3": init.InitializeF3NCObjectsZM,
        "g1": init.Initializeg1NCObjectsZM,
        "gL": init.InitializegLNCObjectsZM,
        "g4": init.Initializeg4NCObjectsZM,
    }

    MAP_FFNS_NC = {
        "F2": init.InitializeF2NCObjectsMassive,
        "FL": init.InitializeFLNCObjectsMassive,
    }
    MAP_FFN0_NC = {
        "F2": init.InitializeF2NCObjectsMassiveZero,
        "FL": init.InitializeFLNCObjectsMassiveZero,
    }

    if observables["prDIS"] == "CC":
        if PROJECTILE_PIDS[observables["ProjectileDIS"]] > 0:
            return MAP_ZM_CC
        raise ValueError(f"{observables['ProjectileDIS']} not available in Apfel++")

    if fns == "ZM-VFNS":
        return MAP_ZM_NC
    if fns == "FFNS":
        return MAP_FFNS_NC
    if fns == "FFN0":
        return MAP_FFN0_NC


def couplings(ap, pids, proc_type, obs_name):
    """Return the corresponding coupling given a process type and an observable.

    Parameters
    ----------
    pids: int
        PDG ID of the projectile
    proc_type:
        type of the DIS process
    obs_name: str
        name of the Yadism observable

    Returns
    -------
    callable
        a callable function that computes the coupling as a
        function of the scale Q

    """

    def _fBq(Q):
        if proc_type == "EM":
            # For Q=0 we only have electric charges
            return ap.utilities.ElectroWeakCharges(0, False, pids)
        return ap.utilities.ElectroWeakCharges(Q, False, pids)

    def _fDq(Q):
        if proc_type == "EM":
            return [0.0]
        return ap.utilities.ParityViolatingElectroWeakCharges(Q, False, pids)

    # CKM matrix elements
    def _fCKM(Q):
        return ap.constants.CKM2

    if proc_type == "CC":
        coupling = _fCKM
        if obs_name.startswith("g"):
            raise ValueError("APFEL++ cannot compute polarised CC yet.")
    else:
        coupling = _fBq
        if "F3" in obs_name or "gL" in obs_name or "g4" in obs_name:
            coupling = _fDq

    return coupling


def tabulate_nc(ap, obs_name, sfobj, nq, qmin, qmax, thrs):
    """Tabulate the NC/EM structure function predictions.

    Parameters
    ----------
    obs_name: str
        name of the observable
    sfobj: ap.init.InitalizeSFNCOjbectsZM
        SF objects in Zero-Mass Flavour Number Scheme
    nq: int
        number of Q points
    qmin: float
        minimal value of Q
    qmax: float
        maximal value of Q
    thrs: list
        list of quark masses and thresholds

    Returns
    -------
    callable:
        Apfel++ callable functions to evaluate SF from
    """
    if "total" in obs_name:
        tab_sf = ap.TabulateObjectD(sfobj[0].Evaluate, nq, qmin, qmax, 3, thrs)
    elif "light" in obs_name:
        tab_sf = ap.TabulateObjectD(
            lambda Q: sfobj[1].Evaluate(Q)
            + sfobj[2].Evaluate(Q)
            + sfobj[3].Evaluate(Q),
            nq,
            qmin,
            qmax,
            3,
            thrs,
        )
    elif "charm" in obs_name:
        tab_sf = ap.TabulateObjectD(sfobj[4].Evaluate, nq, qmin, qmax, 3, thrs)
    elif "bottom" in obs_name:
        tab_sf = ap.TabulateObjectD(sfobj[5].Evaluate, nq, qmin, qmax, 3, thrs)
    else:
        raise ValueError(f"'(total, light, charm, bottom)' not found in {obs_name}!")
    return tab_sf


def tabulate_cc(ap, obs_name, sfobj, nq, qmin, qmax, thrs):
    """Tabulate the CC structure function predictions.

    Parameters
    ----------
    obs_name: str
        name of the observable
    sfobj: list(ap.init.InitalizeSFNCOjbectsZM)
        list of SF objects in Zero-Mass Flavour Number Scheme
    nq: int
        number of Q points
    qmin: float
        minimal value of Q
    qmax: float
        maximal value of Q
    thrs: list
        list of quark masses and thresholds

    Returns
    -------
    callable:
        Apfel++ callable functions to evaluate SF from
    """
    sfp, sfm = sfobj
    if "total" in obs_name:
        tab_sf = ap.TabulateObjectD(
            lambda Q: 2 * (sfp[0].Evaluate(Q) - sfm[0].Evaluate(Q)),
            nq,
            qmin,
            qmax,
            3,
            thrs,
        )
    elif "light" in obs_name:
        tab_sf = ap.TabulateObjectD(
            lambda Q: 2
            * (
                sfp[1].Evaluate(Q)
                + sfp[2].Evaluate(Q)
                - (sfm[1].Evaluate(Q) + sfm[2].Evaluate(Q))
            ),
            nq,
            qmin,
            qmax,
            3,
            thrs,
        )
    elif "charm" in obs_name:
        tab_sf = ap.TabulateObjectD(
            lambda Q: 2
            * (
                sfp[4].Evaluate(Q)
                + sfp[5].Evaluate(Q)
                - (sfm[4].Evaluate(Q) + sfm[5].Evaluate(Q))
            ),
            nq,
            qmin,
            qmax,
            3,
            thrs,
        )
    elif "bottom" in obs_name:
        tab_sf = ap.TabulateObjectD(
            lambda Q: 2
            * (
                sfp[3].Evaluate(Q)
                + sfp[6].Evaluate(Q)
                - (sfm[3].Evaluate(Q) + sfm[6].Evaluate(Q))
            ),
            nq,
            qmin,
            qmax,
            3,
            thrs,
        )
    else:
        raise ValueError(f"'(total, light, charm, bottom)' not found in {obs_name}!")
    return tab_sf


def compute_apfelpy_data(theory, observables, pdf):
    """Run APFEL++ to compute observables.

    Parameters
    ----------
    theory : dict
        theory runcard
    observables : dict
        observables runcard
    pdf : lahapdf_like
        PDF set

    Returns
    -------
    dict
        AFPEL numbers
    """
    import apfelpy as ap  # pylint: disable=import-error, import-outside-toplevel

    ap.Banner()

    # Setup APFEL x-grid
    xgrid = ap.Grid(
        [
            ap.SubGrid(100, 1e-5, 3),
            ap.SubGrid(60, 1e-1, 3),
            ap.SubGrid(50, 6e-1, 3),
            ap.SubGrid(50, 8e-1, 3),
        ]
    )

    # Here we multiply the `thr` scales
    thrs = [
        0,
        0,
        0,
        theory["mc"] * theory["kcThr"],
        theory["mb"] * theory["kbThr"],
        theory["mt"] * theory["ktThr"],
    ]
    masses = [
        0,
        0,
        0,
        theory["mc"],
        theory["mb"],
        theory["mt"],
    ]

    # Setting the theory
    fns = theory["FNS"]
    if fns not in ["ZM-VFNS", "FFNS", "FFN0"]:
        raise ValueError(f"APFEL++ does not contain {fns}.")

    # Perturbative Order
    pto = theory["PTO"]

    # Couplings
    alphas = ap.AlphaQCD(theory["alphas"], theory["Qref"], masses, thrs, pto)
    alphas = ap.TabulateObject(alphas, 2 * NQ, QMIN, QMAX, 3)

    # Map Yadism observables to Apfel++ Objects
    apfelpy_structure_functions = map_apfelpy_sf(ap.initializers, observables, fns)

    # Initialize DGLAP Object
    dglapobj = ap.initializers.InitializeDglapObjectsQCD(xgrid, masses, thrs)

    apf_tab = {}
    for obs_name, kinematics in observables["observables"].items():
        apf_tab[obs_name] = []

        sf_name, heaviness = obs_name.split("_")
        if fns != "ZM-VFNS" and heaviness == "total":
            raise ValueError(
                "total is not provided in APFEL++ for massive calculations."
            )
        pids = MAP_HEAVINESS[heaviness]

        # Define the couplings
        coupling = couplings(ap, pids, observables["prDIS"], obs_name)

        # iterate over input kinematics
        for kin in kinematics:
            Q2 = kin["Q2"]
            x = kin["x"]

            # Construct the DGLAP objects
            if "ToyLH" in pdf.set().name:
                pdf_xfxQ = lambda x, mu: ap.utilities.PhysToQCDEv(
                    {pid: pdf.xfxQ(pid, x, mu) for pid in br.flavor_basis_pids}
                )
            else:
                pdf_xfxQ = lambda x, mu: ap.utilities.PhysToQCDEv(pdf.xfxQ(x, mu))
            evolvedPDFs = ap.builders.BuildDglap(
                dglapobj,
                pdf_xfxQ,
                np.sqrt(Q2) * theory["XIF"],
                pto,
                alphas.Evaluate,
            )
            # Tabulate PDFs
            tabulatedPDFs = ap.TabulateObjectSetD(evolvedPDFs, NQ, QMIN, QMAX, 3)

            # Initialize structure functions
            eff_thrs = thrs if fns == "ZM_VFNS" else masses
            if observables["prDIS"] == "CC":
                sfobj = []
                for sign in ["p", "m"]:
                    tmp_sf = apfelpy_structure_functions[f"{sf_name}{sign}"](
                        xgrid, eff_thrs
                    )
                    sfobj.append(
                        ap.builders.BuildStructureFunctions(
                            tmp_sf,
                            tabulatedPDFs.EvaluateMapxQ,
                            pto,
                            alphas.Evaluate,
                            coupling,
                            theory["XIR"],
                            theory["XIF"],
                        )
                    )
                tab_sf = tabulate_cc(ap, obs_name, sfobj, NQ, QMIN, QMAX, thrs)
            else:
                sfobj = apfelpy_structure_functions[sf_name](xgrid, eff_thrs)
                sfobj = ap.builders.BuildStructureFunctions(
                    sfobj,
                    tabulatedPDFs.EvaluateMapxQ,
                    pto,
                    alphas.Evaluate,
                    coupling,
                    theory["XIR"],
                    theory["XIF"],
                )
                tab_sf = tabulate_nc(ap, obs_name, sfobj, NQ, QMIN, QMAX, thrs)

            # shift convolution for massive
            x_eval = x
            if fns == "FFNS" and heaviness != "light":
                m_h = masses[MAP_HEAVINESS[heaviness] - 1]
                eta = Q2 / (Q2 + 4 * m_h**2)
                x_eval = x / eta
            # compute the actual result
            result = tab_sf.EvaluatexQ(x_eval, np.sqrt(Q2))

            # take over the kinematics
            r = kin.copy()
            r["result"] = result
            apf_tab[obs_name].append(r)

    return apf_tab



"""Note, the factor `fl11` is disabled for charged currents according to `xcdiff3p.f`.
To generate `clnm3a_fl2,clnm3c_fl2` we follow `xcdiff3p.f` or the reference paper :cite:`Davies:2016ruz`
"""

import numba as nb
import numpy as np

from ...special.nielsen import nielsen
from ...special.zeta import zeta2
from .common import d81
from .xcdiff3p import clq3dfp


@nb.njit("f8(f8,f8[:])", cache=True)
def clnp3a_fl2(y, args):
    """The math:`fl_{2}`: regular piece of the non singlet coefficient."""
    nf = args[0]
    dl = np.log(y)
    dl1 = np.log(1.0 - y)
    li2 = nielsen(1, 1, y).real
    res = (
        -2220.5
        - 7884.0 * y
        + 4168.0 * y**2
        - 1280.0 * d81 * dl**3
        - 7456.0 / 27.0 * dl**2
        - 1355.7 * dl
        + 512.0 / 27 * dl1**4
        - 177.40 * dl1**3
        + 650.6 * dl1**2
        - 2729.0 * dl1
        + 208.3 * y * dl**3
        - dl1**3 * (1.0 - y) * (125.3 - 195.6 * dl1)
        - dl * dl1 * (844.7 * dl + 517.3 * dl1)
        + nf
        * (
            408.4
            - 9.345 * y
            - 919.3 * y**2
            + 1728.0 * d81 * dl**2
            + 200.73 * dl
            - 1792.0 * d81 * y * dl**3
            + 1024.0 * d81 * dl1**3
            - 112.35 * dl1**2
            + 344.1 * dl1
            + (1.0 - y) * dl1**2 * (239.7 + 20.63 * dl1)
            + dl * dl1 * (887.3 + 294.5 * dl - 59.14 * dl1)
        )
        + nf**2
        * (
            -19.0
            + (317.0 / 6.0 - 12.0 * zeta2) * y
            + 9.0 * y * dl**2
            + dl * (-6.0 + 50.0 * y)
            + 3.0 * y * dl1**2
            + dl1 * (6.0 - 25.0 * y)
            - 6.0 * y * dl * dl1
            + 6.0 * y * li2
        )
        * 64.0
        * d81
    )
    return res


@nb.njit("f8(f8,f8[:])", cache=True)
def clnp3c_fl2(y, args):
    """The math:`fl_{2}`: local piece of the non singlet coefficient."""
    nf = args[0]
    res = 0.113 + nf * 0.006
    return res


@nb.njit("f8(f8,f8[:])", cache=True)
def clnm3a_fl2(y, args):
    return clnp3a_fl2(y, args) - clq3dfp(y, args)


@nb.njit("f8(f8,f8[:])", cache=True)
def clnm3c_fl2(y, args):
    return clnp3c_fl2(y, args)



import numpy as np
from eko import constants
from scipy.special import spence

from .. import partonic_channel as pc
from .. import splitting_functions as split


class NeutralCurrentBase(pc.PartonicChannel):
    """
    Heavy partonic coefficient functions that respect hadronic and partonic
    thresholds.
    """

    def __init__(self, *args, m2hq, n3lo_cf_variation=0):
        self.m2hq = m2hq
        super().__init__(*args)
        # FH - Vogt comparison prefactor
        self._FHprefactor = self.ESF.Q2 / (np.pi * m2hq)

        # common variables
        self._xi = self.ESF.Q2 / m2hq
        self._eta = lambda z: self._xi / 4.0 * (1.0 / z - 1.0) - 1.0
        self.n3lo_cf_variation = n3lo_cf_variation

    def decorator(self, f):
        """
        Apply hadronic threshold

        Parameters
        ----------
            f : callable
                input

        Returns
        -------
            f : callable
                output
        """
        if self.is_below_pair_threshold(self.ESF.x):
            return lambda: pc.RSL()
        return f

    def is_below_pair_threshold(self, z):
        """
        Checks if the available energy is below production threshold or not

        Parameters
        ----------
            z : float
                partonic momentum fraction

        Returns
        -------
            is_below_pair_threshold : bool
                is the partonic energy sufficient to create the heavy quark
                pair?

        .. todo::
            use threshold on shat or using FH's zmax?
        """
        shat = self.ESF.Q2 * (1 - z) / z
        return shat <= 4 * self.m2hq


class ChargedCurrentBase(pc.PartonicChannel):
    r"""
    Heavy partonic coefficient functions that respect hadronic and partonic
    thresholds.

    From :cite:`gluck-ccheavy` we see that the partonic coefficient functions have to be multiplied
    by different factors for the different structure functions. In order to keep
    track of this we use the :attr:`sf_prefactor` attribute. Coefficients
    that to not explicitly depend on the structure function kind are mulitplied
    by this factor (:math:`B_{3,i}` and the explicit factorization bits).
    For all the other the factors have to be applied explicitly: e.g.
    :math:`A_L = A_2 - \lambda A_1`.

    Parameters
    ----------
        m2hq : float
            heavy quark mass
    """

    def __init__(self, *args, m2hq):
        super().__init__(*args)
        # common variables
        self.labda = 1.0 / (1.0 + m2hq / self.ESF.Q2)
        self.x = self.ESF.x
        self.ka = 1.0 / self.labda * (1.0 - self.labda) * np.log(1.0 - self.labda)
        self.l_labda = lambda z, labda=self.labda: np.log(
            (1.0 - labda * z) / (1.0 - labda) / z
        )
        # normalization helper
        self.sf_prefactor = 1.0

    def convolution_point(self):
        """
        Change convolution point due to massive particles
        """
        return self.x / self.labda


class ChargedCurrentNonSinglet(ChargedCurrentBase):
    """Quark contributions"""

    def r_integral(self, x):
        r"""
        -Power(Pi,2)/6. + Li2(\[Lambda]) + Li2((1 - \[Lambda])/(1 -
        x*\[Lambda])) + ln(1 - \[Lambda])*ln(\[Lambda]) - ln(1 - x)*ln(1 -
        x*\[Lambda]) - ln(\[Lambda])*ln(1 - x*\[Lambda]) + Power(ln(1 -
        x*\[Lambda]),2)/2.
        """
        labda = self.labda
        return (
            -np.pi**2 / 6
            + spence(1 - labda)
            + spence((1 - x) * labda / (1 - x * labda))
            + np.log(1 - labda) * np.log(labda)
            - np.log(1 - x) * np.log(1 - x * labda)
            - np.log(labda) * np.log(1 - x * labda)
            + np.log(1 - x * labda) ** 2 / 2
        )

    def h_q(self, a, b1, b2):
        CF = constants.CF

        b3 = self.sf_prefactor / 2
        as_norm = 2

        def reg(z, _args):
            hq_reg = -(1 + z**2) * np.log(z) / (1 - z) - (1 + z) * (
                2 * np.log(1 - z) - np.log(1 - self.labda * z)
            )
            return (
                (
                    -self.sf_prefactor
                    * np.log(self.labda)
                    * (split.lo.pqq_reg(z, np.array([], dtype=float)) / 2.0)
                )
                + CF
                * (
                    self.sf_prefactor * hq_reg
                    + (b1(z) - b1(1)) / (1 - z)
                    + (b2(z) - b2(1)) / (1 - self.labda * z)
                    # b3(z) - b3(1) = 0 = 1/2 - 1/2
                )
            ) * as_norm

        def sing(z, _args):
            hq_sing = 2 * ((2 * np.log(1 - z) - np.log(1 - self.labda * z)) / (1 - z))
            return (
                (
                    -self.sf_prefactor
                    * np.log(self.labda)
                    * (split.lo.pqq_sing(z, np.array([], dtype=float)) / 2)
                )
                + CF
                * (
                    self.sf_prefactor * hq_sing
                    + b1(1) / (1 - z)
                    + b2(1) / (1 - self.labda * z)
                    + b3 * (1 - z) / (1 - self.labda * z) ** 2
                )
            ) * as_norm

        def local(x, _args):
            log_pd_int = -np.log(1 - x) ** 2 / 2.0
            hq_loc = -(
                4.0
                + 1.0 / (2.0 * self.labda)
                + np.pi**2 / 3.0  # see erratum
                + (1.0 + 3.0 * self.labda) / (2.0 * self.labda) * self.ka
            ) - 2.0 * (2.0 * log_pd_int - self.r_integral(x))

            b1_int = -np.log(1.0 - x)

            b2_int = -np.log(1.0 - x * self.labda) / self.labda

            b3_int = (
                -((x * (-1.0 + self.labda)) / (self.labda * (-1.0 + x * self.labda)))
                - np.log(1.0 - self.labda * x) / self.labda**2
            )

            return (
                (
                    -self.sf_prefactor
                    * np.log(self.labda)
                    * (split.lo.pqq_local(x, np.array([], dtype=float)) / 2.0)
                )
                + CF
                * (
                    self.sf_prefactor * hq_loc
                    + a
                    - b1(1) * b1_int
                    - b2(1) * b2_int
                    - b3 * b3_int
                )
            ) * as_norm

        return pc.RSL(reg, sing, local)

    def LO(self):
        return pc.RSL.from_delta(self.sf_prefactor)


class ChargedCurrentGluon(ChargedCurrentBase):
    """Gluon contributions"""

    def h_g(self, z, cs):
        c0 = (
            self.sf_prefactor
            * (split.lo.pqg_single(z, np.array([], dtype=float)) / 2.0)
            * (2.0 * np.log(1 - z) - np.log(1 - self.labda * z) - np.log(z))
        )
        cs.insert(0, c0)
        return (
            cs[0]
            + cs[1] * z * (1 - z)
            + cs[2]
            + (cs[3] + self.labda * z * cs[4]) * (1 - self.labda) * z * self.l_labda(z)
        )



"""Output related utilities.

For the main output (that is the computed |PDF|
independent |DIS| operator) three outputs are provided:

- tar archive, containing metadata and binary :mod:`numpy.lib.format` arrays
  (this is the **suggested** output format)
- single file `yaml <https://yaml.org/>`_ output: a single human readable (but
  possibly huge) file
- `PineAPPL <https://github.com/N3PDF/pineappl>`_ interpolation grid: very
  useful to store in a standard format (supporting also non-|DIS| processes) and
  interfacing to other codes (but *no loading* is supported from this format)

"""

import copy
import pathlib
import tarfile
import tempfile

import numpy as np
import pandas as pd
import yaml
from eko.couplings import Couplings, couplings_mod_ev
from eko.io import dictlike, runcards, types
from eko.matchings import Atlas, nf_default
from eko.quantities.heavy_quarks import MatchingScales

from . import observable_name as on
from .esf.result import ESFResult, EXSResult


class MaskedPDF:
    """Mask some pids of a PDF set to be 0.

    Parameters
    ----------
    lhapdf_like : callable
        object that provides an xfxQ2 callable (as `lhapdf
        <https://lhapdf.hepforge.org/>`_ and :class:`ekomark.toyLH.toyPDF` do)
        (and thus is in flavor basis)
    active_pids : list[int]
        active PIDs

    """

    def __init__(self, lhapdf_like, active_pids):
        self.parent = lhapdf_like
        self.active_pids = active_pids

    def __getattr__(self, name):
        return self.parent.__getattribute__(name)

    def xfxQ2(self, pid, x, Q2):
        """Fake lhapdf-like wrapper."""
        return self.parent.xfxQ2(pid, x, Q2) if pid in self.active_pids else 0.0


class Output(dict):
    """Wrapper for the output to help with application to PDFs and dumping to file."""

    theory = None
    observables = None

    def apply_pdf(self, lhapdf_like):
        r"""Compute all observables for the given PDF.

        Parameters
        ----------
        lhapdf_like : object
            object that provides an xfxQ2 callable (as `lhapdf
            <https://lhapdf.hepforge.org/>`_ and :class:`ekomark.toyLH.toyPDF`
            do) (and thus is in flavor basis)

        Returns
        -------
        ret : :class:`PDFOutput`
            output dictionary with all structure functions for all x, Q2, result and error

        """
        return self.apply_pdf_theory(lhapdf_like, self.theory)

    def apply_pdf_theory(self, lhapdf_like, theory):
        r"""Compute all observables for the given PDF.

        Parameters
        ----------
        lhapdf_like : object
            object that provides an xfxQ2 callable (as `lhapdf
            <https://lhapdf.hepforge.org/>`_ and :class:`ekomark.toyLH.toyPDF`
            do) (and thus is in flavor basis)
        theory : dict
            theory dictionary

        Returns
        -------
        ret : :class:`PDFOutput`
            output dictionary with all structure functions for all x, Q2, result and error

        """
        new_eko_theory = runcards.Legacy(theory=theory, operator={}).new_theory
        method = runcards.Legacy.MOD_EV2METHOD.get(theory["ModEv"], theory["ModEv"])
        method = dictlike.load_enum(types.EvolutionMethod, method)
        method = couplings_mod_ev(method)
        masses = [mq**2 for mq, _ in new_eko_theory.heavy.masses]
        thresholds_ratios = np.power(new_eko_theory.heavy.matching_ratios, 2)
        sc = Couplings(
            couplings=new_eko_theory.couplings,
            order=new_eko_theory.order,
            method=method,
            masses=masses,
            hqm_scheme=new_eko_theory.heavy.masses_scheme,
            thresholds_ratios=thresholds_ratios.tolist(),
        )
        atlas = Atlas(
            matching_scales=MatchingScales(masses * thresholds_ratios),
            origin=(theory["Qref"] ** 2, theory["nfref"]),
        )
        fns = theory["FNS"]
        if "FFNS" in fns or "FFN0" in fns:
            alpha_s = lambda muR: sc.a_s(muR**2, nf_to=theory["NfFF"]) * 4.0 * np.pi
        elif fns == "ZM-VFNS":
            alpha_s = (
                lambda muR: sc.a_s(muR**2, nf_to=nf_default(muR**2, atlas))
                * 4.0
                * np.pi
            )
        else:
            raise ValueError(f"Scheme '{fns}' not recognized.")
        alpha_qed = lambda _muR: theory["alphaqed"]
        return self.apply_pdf_alphas_alphaqed_xir_xif(
            lhapdf_like, alpha_s, alpha_qed, theory["XIR"], theory["XIF"]
        )

    def apply_pdf_alphas_alphaqed_xir_xif(
        self, lhapdf_like, alpha_s, alpha_qed, xiR, xiF
    ):
        r"""Compute all observables for the given PDF.

        Parameters
        ----------
        lhapdf_like : object
            object that provides an xfxQ2 callable (as `lhapdf
            <https://lhapdf.hepforge.org/>`_ and :class:`ekomark.toyLH.toyPDF`
            do) (and thus is in flavor basis)
        alpha_s : callable
            :math:`\alpha_s(\mu_R)`, the running strong coupling
        alpha_qed : callable
            :math:`\alpha(\mu_R)`, the running fine structure constant
        xiR : float
            ratio renormalization scale to |EW| boson virtuality (linear!)
        xiF : float
            ratio factorization scale to |EW| boson virtuality (linear!)

        Returns
        -------
        ret : :class:`PDFOutput`
            output dictionary with all structure functions for all :math:`x`,
            :math:`Q^2`, result and error

        """
        # iterate
        ret = PDFOutput()

        xgrid = self["xgrid"]["grid"]

        # dispatch onto result
        for obs in self:
            if not on.ObservableName.is_valid(obs):
                continue
            if self[obs] is None:
                continue
            ret[obs] = []
            for kin in self[obs]:
                ret[obs].append(
                    kin.apply_pdf(
                        lhapdf_like, self["pids"], xgrid, alpha_s, alpha_qed, xiR, xiF
                    )
                )
        return ret

    def get_raw(self):
        """Serialize result as dict.

        This maps the original numpy matrices to lists.

        Returns
        -------
        out : dict
            dictionary which will be written on output

        """
        out = {}
        # dump raw elements
        for f in self:
            out[f] = copy.copy(self[f])
        out["pids"] = list(self["pids"])
        # make raw lists
        for k in ["xgrid"]:
            out[k]["grid"] = self[k]["grid"]
            out[k]["log"] = self[k]["log"]
        for obs in self:
            if not on.ObservableName.is_valid(obs):
                continue
            if self[obs] is None:
                continue
            out[obs] = []
            for kin in self[obs]:
                out[obs].append(kin.get_raw())
        return out

    def dump_yaml(self, stream=None):
        """Serialize result as YAML.

        Parameters
        ----------
        stream : None or stream
            if given, dump is written on it

        Returns
        -------
        dump : any
            result of dump(output, stream), i.e. a string, if no stream is given or
            Null, if self is written sucessfully to stream

        """
        # TODO explicitly silence yaml
        out = self.get_raw()
        out["theory"] = self.theory
        out["observables"] = self.observables
        return yaml.dump(out, stream, default_flow_style=None)

    def dump_yaml_to_file(self, filename):
        """Write YAML representation to a file.

        Parameters
        ----------
        filename : str or os.PathLike
            target file name

        Returns
        -------
        ret : any
            result of dump(output, stream), i.e. Null if written sucessfully

        """
        with open(filename, "w", encoding="utf8") as f:
            ret = self.dump_yaml(f)
        return ret

    def dump_tar(self, tarpath, runcards=True):
        """Serialize output in a tar archive.

        This is the favorite *native* output.

        This results in a considerably smaller output file, with respect to the
        'yaml' serialization, but it looses the readability.
        In most cases, there is no advantage in having a readable DIS operator,
        so they are serialized in binary arrays (see :mod:`numpy.lib.format`).

        This **only supports** observables with a uniform order across
        kinematical point (as they are generated by
        :class:`yadism.runner.Runner`), since they are stored in a unique array.
        They don't have to be uniform across different observables (since they
        are stored in different arrays anyhow).

        Parameters
        ----------
        tarpath : str or os.PathLike
            target file path (it has to have the '.tar' extension)

        Raises
        ------
        AssertionError
            If orders are not uniform within a single observable.

        """
        tarpath = pathlib.Path(tarpath)
        if tarpath.suffix != ".tar":
            raise ValueError(f"'{tarpath}' is not a valid tar filename, wrong suffix")

        with tempfile.TemporaryDirectory() as tmpdir:
            tmpdir = pathlib.Path(tmpdir)

            metadata = {}
            for metafield, metavalue in self.items():
                if not on.ObservableName.is_valid(metafield) or metavalue is None:
                    # this way even scalar types are properly casted
                    # it works for sure for: None, bool, int, float, str
                    # and so everything we need (what is safe in yaml)
                    metadata[metafield] = np.array(metavalue).tolist()
                else:
                    kinematics = {}
                    for key in metavalue[0].get_raw():
                        if key != "orders":
                            kinematics[key] = []

                    orders_first = []
                    values = []
                    errors = []
                    for esfres in metavalue:
                        orders = []
                        esf_values = []
                        esf_errors = []
                        for key, value in esfres.get_raw().items():
                            if key != "orders":
                                kinematics[key].append(value)
                            else:
                                for order in value:
                                    orders.append(order["order"])
                                    esf_values.append(order["values"])
                                    esf_errors.append(order["errors"])

                        if len(orders_first) == 0:
                            orders_first = orders
                        else:
                            assert orders_first == orders
                        values.append(esf_values)
                        errors.append(esf_errors)

                    np.savez_compressed(
                        tmpdir / metafield,
                        values=np.array(values),
                        errors=np.array(errors),
                    )
                    metadata[metafield] = dict(
                        orders=orders_first, kinematics=kinematics
                    )

            (tmpdir / "metadata.yaml").write_text(
                yaml.safe_dump(metadata, default_flow_style=None), encoding="utf-8"
            )
            if runcards:
                runcards = tmpdir / "runcards"
                runcards.mkdir()
                (runcards / "theory.yaml").write_text(
                    yaml.safe_dump(self.theory), encoding="utf-8"
                )
                (runcards / "observables.yaml").write_text(
                    yaml.safe_dump(self.observables, default_flow_style=None),
                    encoding="utf-8",
                )

            with tarfile.open(tarpath, "w") as tar:
                tar.add(tmpdir, arcname=tarpath.stem)

    @classmethod
    def load_tar(cls, tarpath):
        """Deserialize output object from tar.

        It loads an :meth:`Output.dump_tar` generated tar file into an
        :class:`Output` object.

        Parameters
        ----------
        tarpath : str or os.PathLike
            target file path (it has to be a 'tar' archive)

        Returns
        -------
        :class:`Output`
            loaded object

        """
        tarpath = pathlib.Path(tarpath)

        # Initialize output object
        out = cls()

        with tempfile.TemporaryDirectory() as tmpdir:
            tmpdir = pathlib.Path(tmpdir)

            with tarfile.open(tarpath, "r") as tar:
                tar.extractall(tmpdir)

            # The internal name is the one that has been used to save, since the
            # tar could have been renamed in the meanwhile, it might be unknown,
            # but there is only a single folder
            innerdir = list(tmpdir.glob("*"))[0]

            runcards = innerdir / "runcards"
            out.theory = yaml.safe_load((runcards / "theory.yaml").read_text())
            out.observables = yaml.safe_load(
                (runcards / "observables.yaml").read_text()
            )

            metadata = yaml.safe_load((innerdir / "metadata.yaml").read_text())
            for metafield, metavalue in metadata.items():
                if not on.ObservableName.is_valid(metafield) or metavalue is None:
                    ar = np.array(metavalue)
                    if ar.ndim > 0:
                        out[metafield] = ar
                    else:
                        out[metafield] = metavalue
                else:
                    op = np.load(innerdir / f"{metafield}.npz")
                    kinvars, kinvalues = list(zip(*metavalue["kinematics"].items()))
                    kinematics = [dict(zip(kinvars, kin)) for kin in zip(*kinvalues)]
                    orders = metavalue["orders"]

                    Result = ESFResult if "y" not in kinvars else EXSResult
                    results = []

                    for kin, val, err in zip(kinematics, op["values"], op["errors"]):
                        res = kin.copy()
                        res["orders"] = [
                            dict(zip(["order", "values", "errors"], ordvalerr))
                            for ordvalerr in zip(orders, val, err)
                        ]
                        results.append(Result.from_document(res))

                    out[metafield] = results

        return out

    @classmethod
    def load_yaml(cls, stream):
        """Load YAML representation from stream.

        Parameters
        ----------
        stream : any
            source stream

        Returns
        -------
        obj : :class:`Output`
            loaded object

        """
        obj = yaml.safe_load(stream)
        # make list numpy
        for k in ["xgrid"]:
            obj[k]["grid"] = np.array(obj[k]["grid"])
            obj[k]["log"] = obj[k]["log"]
        for obs in obj:
            if not on.ObservableName.is_valid(obs):
                continue
            if obj[obs] is None:
                continue
            Result = ESFResult if "y" not in obj[obs][0] else EXSResult
            for j, kin in enumerate(obj[obs]):
                obj[obs][j] = Result.from_document(kin)

        out = cls(obj)
        out.theory = obj["theory"]
        out.observables = obj["observables"]
        del out["theory"]
        del out["observables"]
        return out

    @classmethod
    def load_yaml_from_file(cls, filename):
        """Load YAML representation from file.

        Parameters
        ----------
        filename : str or os.PathLike
            source file name

        Returns
        -------
        obj : :class:`Output`
            loaded object

        """
        obj = None
        with open(filename) as o:
            obj = cls.load_yaml(o)
        return obj


class PDFOutput(Output):
    """Wrapper for the PDF output to help with dumping to file."""

    def get_raw(self):
        """Convert the object into a native Python dictionary.

        Returns
        -------
        out : dict
            raw dictionary

        """
        out = {}
        for obs in self:
            if self[obs] is None:
                continue
            out[obs] = []
            for kin in self[obs]:
                out[obs].append({k: float(v) for k, v in kin.items()})
        return out

    @classmethod
    def load_yaml(cls, stream):
        """Load the object from YAML.

        Parameters
        ----------
        stream : any
            source stream

        Returns
        -------
        obj : :class:`PDFOutput`
            loaded object

        """
        obj = yaml.safe_load(stream)
        return cls(obj)

    @property
    def tables(self):
        """Convert data into a mapping structure functions -> :class:`pandas.DataFrame`."""
        tables = {}
        for k, v in self.items():
            tables[k] = pd.DataFrame(v)

        return tables

    def dump_tables_to_file(self, filename):
        """Write all tables to file.

        Parameters
        ----------
        filename : str
            output file name

        """
        with open(filename, "w") as f:
            for name, table in self.tables.items():
                f.write("\n".join([name, str(table), "\n"]))

    def dump_tables_to_csv(self, dirname):
        """Write all tables to separate csv files.

        Parameters
        ----------
        dirname : str
            output directory name

        """
        dirname = pathlib.Path(dirname)
        dirname.mkdir(exist_ok=True)
        for name, table in self.tables.items():
            filename = dirname / f"{name}.csv"
            table.to_csv(filename)



from ..partonic_channel import RSL
from . import f2_nc, n3lo, nnlo


class NonSingletEven(f2_nc.NonSinglet):
    pass


class Gluon(f2_nc.Gluon):
    pass


class Singlet(f2_nc.Singlet):
    pass


class NonSingletOdd(f2_nc.NonSinglet):
    def NNLO(self):
        """
        |ref| implements :eqref:`2.9`, :cite:`vogt-f2lcc`.
        """

        return RSL(
            nnlo.xc2ns2p.c2nc2a, nnlo.xc2ns2p.c2ns2b, nnlo.xc2ns2p.c2nc2c, [self.nf]
        )

    def N3LO(self):
        """
        |ref| implements the difference between :eqref:`2.9`, and :eqref:`3.6`, :cite:`Davies:2016ruz`
        """

        return RSL(
            n3lo.xc2ns3p.c2nm3a_fl2,
            n3lo.xc2ns3p.c2ns3b_fl2,
            n3lo.xc2ns3p.c2nm3c_fl2,
            [self.nf],
        )



from ..partonic_channel import RSL
from . import partonic_channel as pc


class Rplus(pc.NeutralCurrentBase):
    def LO(self):
        factor = self.ESF.x / self.convolution_point()
        return RSL.from_delta(factor)

    def NLO(self):
        return self.mkNLO("3", "rplus")


class Rminus(pc.NeutralCurrentBase):
    def NLO(self):
        return self.mkNLO("3", "rminus")



"""Localized implementation of BenchmarkRunner."""

import numpy as np
import pandas as pd
from banana.benchmark.runner import BenchmarkRunner
from banana.data import dfdict
from eko.couplings import Couplings, couplings_mod_ev
from eko.io import dictlike, runcards, types
from eko.matchings import Atlas, nf_default
from eko.quantities.heavy_quarks import MatchingScales

import yadism
from yadmark.data import db, observables


class Runner(BenchmarkRunner):
    """Localized implementation of BenchmarkRunner."""

    alphas_from_lhapdf = False
    """Use the alpha_s routine provided by the Pdf?"""

    db_base_cls = db.Base

    @staticmethod
    def load_ocards(conn, ocard_updates):
        """Load observable cards."""
        return observables.load(conn, ocard_updates)

    def run_me(self, theory, ocard, pdf):
        """
        Run yadism.

        Parameters
        ----------
        theory : dict
            theory card
        ocard : dict
            observable card
        pdf : lhapdf_like
            PDF set

        Returns
        -------
        out : yadism.output.PDFOutput
            yadism output

        """
        runner = yadism.Runner(theory, ocard)
        # choose alpha_s source
        if self.alphas_from_lhapdf:
            import lhapdf  # pylint:disable=E0401,C0415

            alpha_s = lambda muR: lhapdf.mkAlphaS(pdf.set().name).alphasQ(muR)
        else:
            new_eko_theory = runcards.Legacy(theory=theory, operator={}).new_theory
            method = runcards.Legacy.MOD_EV2METHOD.get(theory["ModEv"], theory["ModEv"])
            method = dictlike.load_enum(types.EvolutionMethod, method)
            method = couplings_mod_ev(method)
            masses = [mq**2 for mq, _ in new_eko_theory.heavy.masses]
            thresholds_ratios = np.power(new_eko_theory.heavy.matching_ratios, 2)
            sc = Couplings(
                couplings=new_eko_theory.couplings,
                order=new_eko_theory.order,
                method=method,
                masses=masses,
                hqm_scheme=new_eko_theory.heavy.masses_scheme,
                thresholds_ratios=thresholds_ratios.tolist(),
            )
            atlas = Atlas(
                matching_scales=MatchingScales(masses * thresholds_ratios),
                origin=(theory["Qref"] ** 2, theory["nfref"]),
            )
            alpha_s = (
                lambda muR: sc.a_s(muR**2, nf_to=nf_default(muR**2, atlas))
                * 4.0
                * np.pi
            )

        alpha_qed = lambda _muR: theory["alphaqed"]
        return runner.get_result().apply_pdf_alphas_alphaqed_xir_xif(
            pdf, alpha_s, alpha_qed, theory["XIR"], theory["XIF"]
        )

    def run_external(self, theory, ocard, pdf):
        """
        Run external.

        Parameters
        ----------
        theory : dict
            theory card
        ocard : dict
            observable card
        pdf : lhapdf_like
            PDF set

        Returns
        -------
        dict
            external output

        """
        observable = ocard

        if self.external.upper() == "APFEL":
            from .external import (  # pylint:disable=import-error,import-outside-toplevel
                apfel_utils,
            )

            return apfel_utils.compute_apfel_data(theory, observable, pdf)

        elif self.external.upper() == "QCDNUM":
            from .external import (  # pylint:disable=import-error,import-outside-toplevel
                qcdnum_utils,
            )

            return qcdnum_utils.compute_qcdnum_data(theory, observable, pdf)

        elif self.external.lower() == "xspace_bench":
            from .external import (  # pylint:disable=import-error,import-outside-toplevel
                xspace_bench_utils,
            )

            return xspace_bench_utils.compute_xspace_bench_data(theory, observable, pdf)

        elif self.external.upper() == "APFEL++":
            from .external import apfelpy_utils

            return apfelpy_utils.compute_apfelpy_data(theory, observable, pdf)

        elif self.external.lower() == "void":
            # set all ESF simply to 0
            res = {}
            for sf, esfs in ocard["observables"].items():
                if not yadism.observable_name.ObservableName.is_valid(sf):
                    continue
                void_esfs = []
                for esf in esfs:
                    n = esf.copy()
                    n.update({"result": 0})
                    void_esfs.append(n)
                res[sf] = void_esfs
            return res
        raise ValueError("Unknown external")

    def log(self, t, o, _pdf, me, ext):
        """Write log."""
        log_tab = dfdict.DFdict()
        kins = ["x", "Q2"]

        for sf in me:
            if not yadism.observable_name.ObservableName.is_valid(sf):
                continue
            esfs = []

            obs_kins = kins.copy()

            # skip if it is empty
            if len(me[sf]) == 0:
                continue

            if "y" in me[sf][0]:
                obs_kins += ["y"]

            # Sort the point using yadism order since yadism list can be different from ext
            for yad in me[sf]:
                cnt = 0
                for oth in ext[sf]:
                    if all([yad[k] == oth[k] for k in obs_kins]):
                        # add common values
                        esf = {}
                        esf["x"] = yad["x"]
                        esf["Q2"] = yad["Q2"]
                        if "y" in obs_kins:
                            esf["y"] = yad["y"]
                        esf["yadism"] = f = yad["result"]
                        esf["yadism_error"] = yad["error"]
                        esf[self.external] = r = oth["result"]
                        esf["percent_error"] = (f - r) / r * 100
                        esfs.append(esf)
                        cnt = 1
                        break
                if cnt == 0:
                    raise ValueError("Sort problem: x and/or Q2 do not match.")
            log_tab[sf] = pd.DataFrame(esfs)

        return log_tab



# pylint: skip-file
# fmt: off
import numba as nb
import numpy as np
from eko.constants import CA, CF, TR

from ...special import li2


@nb.njit("f8(f8,f8[:])", cache=True)
def pqq0_2_reg(z, _args):
    return -4*CF**2*(5 + z + 4*(1 + z)*np.log(1 - z) - 3*(1 + z)*np.log(z))

@nb.njit("f8(f8,f8[:])", cache=True)
def pqq0_2_sing(z, _args):
    return (-8*CF**2*(3 + 4*np.log(1 - z) - 2*np.log(z)))/(-1 + z)

@nb.njit("f8(f8,f8[:])", cache=True)
def pqq0_2_loc(z, _args):
    return (CF**2*(27 + 8*np.pi**2 + 12*(3 - 4*np.log(1 - z))*np.log(1 - z) - 48*li2(1 - z)))/3.

@nb.njit("f8(f8,f8[:])", cache=True)
def pqq0pqg0_reg(z,args):
    nf = args[0]
    return 4*CF*nf*TR*(-1 + 4*z + (4 + 8*(-1 + z)*z)*np.log(1 - z) + (-2 + 4*z - 8*z**2)*np.log(z))

@nb.njit("f8(f8,f8[:])", cache=True)
def pqg0pgg0_reg(z,args):
    nf = args[0]
    return (4*nf*TR*(-2*nf*z*(1 + 2*(-1 + z)*z) + CA*(8 + z*(17 + 26*z - 40*z**2)) + 12*CA*(z*(1 + 2*(-1 + z)*z)*np.log(1 - z) + z*(1 + 4*z)*np.log(z))))/(3.*z)

@nb.njit("f8(f8,f8[:])", cache=True)
def pqg0pgq0_reg(z,args):
    nf = args[0]
    return 8*CF*nf*TR*(1 + 4/(3.*z) - z - (4*z**2)/3. + 2*(1 + z)*np.log(z))
